import numpy as np
import ds_format as ds
import datetime as dt
from alcf import misc
from alcf.lidars import META

WAVELENGTH = 532 # nm
CALIBRATION_COEFF = 0.375e-5
SURFACE_LIDAR = True
SC_LR = 16.0 # sr
MAX_RANGE = 30000 # m

VARS = {
	'backscatter': ['copol_nrb', 'crosspol_nrb'],
	'backscatter_x': ['copol_nrb', 'crosspol_nrb'],
	'backscatter_y': ['copol_nrb', 'crosspol_nrb'],
}

TIME_VARS = [
	'year',
	'month',
	'day',
	'hour',
	'minute',
	'second',
]

DEFAULT_VARS = TIME_VARS + [
	'range_nrb',
	'elevation_angle',
	'altitude',
	'latitude',
	'longitude',
]

def parse_temporal_resolution(s):
	errmsg = 'Unrecognized temporal resolution "%s"' % s
	x = s.split(' ')
	if len(x) != 2:
		raise ValueError(errmsg)
	FACTOR = [
		[['s', 'sec', 'second', 'seconds'], 1./60./60./24.],
		[['m', 'min', 'minute', 'minutes'], 1./60./24.],
		[['h', 'hour', 'hours'], 1./24.],
	]
	try:
		f = float(x[0])
	except:
		raise ValueError(errmsg)
	for item in FACTOR:
		if x[1] in item[0]:
			return f*item[1]
	raise ValueError(errmsg)

def convert_time(d, tlim):
	time = np.array([
		(
			dt.datetime(int(y), int(m), int(day), int(H), int(M), int(S)) -
			dt.datetime(1970, 1, 1)
		).total_seconds()/(24*60*60) + 2440587.5
		for y, m, day, H, M, S
		in zip(d['year'], d['month'], d['day'], d['hour'], d['minute'], d['second'])
	], np.float64)
	tres = parse_temporal_resolution(d['.']['.']['temporal_resolution'])
	args = [] if tlim is None else [tlim[0], tlim[1]]
	time_bnds = misc.time_bnds(time, tres, *args)
	return time, time_bnds, tres

def read(
	filename,
	vars,
	altitude=None,
	lon=None,
	lat=None,
	tlim=None,
	keep_vars=[],
	**kwargs
):
	sel = None
	tres = None
	if tlim is not None:
		d = ds.read(filename, TIME_VARS)
		d['time'], d['time_bnds'], tres = convert_time(d)
		mask = misc.time_mask(d['time_bnds'], tlim[0], tlim[1])
		if np.sum(mask) == 0: return None
		sel = {'time': mask}

	dep_vars = misc.dep_vars(VARS, vars)
	req_vars = dep_vars + DEFAULT_VARS + keep_vars
	d = ds.read(filename, req_vars, sel=sel, full=True)
	dx = {}
	misc.populate_meta(dx, META, set(vars) & set(VARS))
	n = ds.dim(d, 'time')
	altitude = d['altitude'] if altitude is None else \
		np.full(n, altitude, np.float64)
	lon = d['longitude'] if lon is None else \
		np.full(n, lon, np.float64)
	lat = d['latitude'] if lat is None else \
		np.full(n, lat, np.float64)
	time, time_bnds, tres = convert_time(d, tlim)
	if 'time' in vars:
		dx['time'] = time
		# dx['time'] += 13.0/24.0
	if 'time_bnds' in vars:
		dx['time_bnds'] = time_bnds
	if 'zfull' in vars:
		zfull1 = np.outer(
			np.sin(d['elevation_angle']/180.0*np.pi),
			d['range_nrb']*1e3
		)
		dx['zfull'] = np.tile(zfull1, (n, 1))
		for i in range(n):
			dx['zfull'][i,:] += altitude[i]
	if 'backscatter' in vars:
		dx['backscatter'] = (d['copol_nrb'] + 2.*d['crosspol_nrb'])*CALIBRATION_COEFF
	if 'altitude' in vars:
		dx['altitude'] = altitude
	if 'lon' in vars:
		dx['lon'] = lon
	if 'lat' in vars:
		dx['lat'] = lat
	# if 'backscatter_x' in vars:
	# 	dx['backscatter_x'] = d['copol_nrb']*CALIBRATION_COEFF
	# if 'backscatter_y' in vars:
	# 	dx['backscatter_y'] = d['crosspol_nrb']*CALIBRATION_COEFF
		# 'backscatter_x': {
		# 	'.dims': ['time', 'level'],
		# 	'long_name': 'copolarized_attenuated_backscatter_coefficient',
		# 	'units': 'm-1 sr-1',
		# },
		# 'backscatter_y': {
		# 	'.dims': ['time', 'level'],
		# 	'long_name': 'crosspolarized_attenuated_backscatter_coefficient',
		# 	'units': 'm-1 sr-1',
		# },
	for var in keep_vars:
		misc.keep_var(var, d, dx, {'level': 'range_nrb'})
	return dx



import os
from getpass import getpass

LOGIN_URL = b'https://cds.climate.copernicus.eu/api/v2'

PRODUCTS = ['surf', 'plev']

VARS_SURF = [
	'geopotential',
	'surface_pressure',
]

VARS_PLEV = [
	'fraction_of_cloud_cover',
	'geopotential',
	'specific_cloud_ice_water_content',
	'specific_cloud_liquid_water_content',
	'temperature',
]

TYPE_SURF = 'reanalysis-era5-single-levels'
TYPE_PLEV = 'reanalysis-era5-pressure-levels'

TIME = [
	'00:00', '01:00', '02:00', '03:00', '04:00', '05:00', '06:00', '07:00',
	'08:00', '09:00', '10:00', '11:00',	'12:00', '13:00', '14:00', '15:00',
	'16:00', '17:00', '18:00', '19:00', '20:00', '21:00', '22:00', '23:00'
]

PRESSURE_LEVEL = [
	'1', '2', '3', '5', '7', '10', '20', '30', '50', '70', '100', '125', '150',
	'175', '200', '225', '250', '300', '350', '400', '450', '500', '550', '600',
	'650', '700', '750', '775', '800', '825', '850', '875', '900', '925', '950',
	'975', '1000'
]

def login(uid=None, key=None, overwrite=False):
	if uid is None:
		uid = input('Copernicus CDS UID: ')
	if key is None:
		key = getpass(prompt='Copernicus CDS API key: ')
	if type(uid) is not str:
		uid = str(uid)
	if type(key) is not str:
		key = str(key)

	home = os.path.expanduser("~")
	cdsapirc = os.path.join(home, '.cdsapirc')

	if os.path.exists(cdsapirc) and overwrite is False:
		print('The following files have to be overwritten:')
		print('    ' + cdsapirc)
		res = input('Continue [y/n]? ')
		if res != 'y':
			return

	with open(cdsapirc, 'wb',
		opener=lambda p, f: os.open(p, f, mode=0o600)) as f:
		os.chmod(f.fileno(), 0o600)
		f.write(b'url: %s\n' % LOGIN_URL)
		f.write(b'key: %s:%s\n' % (uid.encode('utf-8'), key.encode('utf-8')))
	print('-> %s' % cdsapirc)

def download(filename, product, year, month, day, lon1, lon2, lat1, lat2,
	nocache=False
):
	import cdsapi
	type_ = {
		'surf': TYPE_SURF,
		'plev': TYPE_PLEV,
	}[product]
	vars_ = {
		'surf': VARS_SURF,
		'plev': VARS_PLEV,
	}[product]
	lon1_180 = lon1 if lon1 < 180 else lon1 - 360
	lon2_180 = lon2 if lon2 < 180 else lon2 - 360
	req = {
			'product_type': 'reanalysis',
			'format': 'netcdf',
			'variable': vars_,
			'time': TIME,
			'area': [lat2, lon1_180, lat1, lon2_180],
			'year': '%d' % year,
			'month': '%d' % month,
			'day': '%d' % day,
	}
	if type_ == 'surf':
		req['pressure_level'] = PRESSURE_LEVEL
	if nocache:
		req['nocache'] = str(random.randint(0, 2147483647))
	c = cdsapi.Client()
	c.retrieve(type_, req, filename)



import sys
import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc
import aquarius_time as aq

VARS = [
	'cli',
	'clw',
	'pfull',
	'ta',
	'ps',
]

STEP = 6/24

def index(dirname, warnings=[], recursive=False, njobs=1):
	dd = ds.readdir(dirname,
		variables=['time'],
		jd=True,
		full=True,
		warnings=warnings,
		recursive=recursive,
		parallel=(njobs > 1),
		njobs=njobs,
	)

	d_g = ds.read(os.path.join(dirname, 'vgrid.nc'), [
		'clon', 'clat'
	], full=True)
	d_g['clon'] *= 180/np.pi
	d_g['clat'] *= 180/np.pi

	return [dd, d_g]

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	vgrid_filename = os.path.join(dirname, 'vgrid.nc')
	dd_out = []
	dd_idx, d_g = index
	ncells = ds.dim(d_g, 'ncells')
	vgrid_cache = {}

	for var in VARS:
		dd = []
		for d_idx in dd_idx:
			if var not in d_idx['.']:
				continue
			time = d_idx['time']
			filename = d_idx['filename']
			ii = np.nonzero(
				(time >= t1 - step*0.5) &
				(time < t2 + step*0.5)
			)[0]
			for i in ii:
				lon0, lat0 = track(time[i])
				if np.isnan(lon0) or np.isnan(lat0):
					continue
				dist = misc.geo_distance(
					np.full(ncells, lon0),
					np.full(ncells, lat0),
					d_g['clon'],
					d_g['clat'],
					method='gc'
				)
				cell = np.argmin(dist)

				if cell in vgrid_cache:
					d_g2 = vgrid_cache[cell]
				else:
					d_g2 = ds.read(vgrid_filename, ['zg', 'zghalf'], sel={
						'ncells': cell,
						'height': ds.dim(d_g, 'height') - 1,
					})
					vgrid_cache[cell] = d_g2

				d = ds.read(filename, [var],
					sel={
						'time': [i],
						'ncells': cell,
						'cell': cell,
					},
					jd=True,
				)
				ds.rename_dim(d, 'height', 'level')
				d['time'] = np.array([time[i]])
				d['lat'] = np.array([d_g['clat'][cell]])
				d['lon'] = np.array([d_g['clon'][cell]])
				d['orog'] = np.array([d_g2['zghalf']])
				d['zfull'] = d_g2['zg'][::-1]
				d['zfull'] = d['zfull'].reshape((1, len(d['zfull'])))
				ds.dims(d, 'time', ['time'])
				ds.dims(d, 'lat', ['time'])
				ds.dims(d, 'lon', ['time'])
				ds.dims(d, 'orog', ['time'])
				ds.dims(d, 'zfull', ['time', 'level'])
				if d[var].ndim == 2:
					d[var] = d[var][:,::-1]
				dd.append(d)
		d = ds.op.merge(dd, 'time')
		dd_out += [d]

	time = set.intersection(*[set(d['time']) for d in dd_out])
	d_out = {}
	for d in dd_out:
		ii = [t in time for t in d['time']]
		ds.select(d, {'time': ii})
		d_out.update(d)

	d_out['cl'] = np.full(d_out['cli'].shape, 100., np.float64)

	return d_out



#!/usr/bin/env python

import sys
import numpy as np
from alcf import misc

def cloud_base_detection(d, **options):
	cloud_mask = d['cloud_mask']
	if len(cloud_mask.shape) == 3:
		n, m, l = d['cloud_mask'].shape
		dims = (n,l)
		dimnames = ('time', 'column')
	else:
		n, m = d['cloud_mask'].shape
		l = 1
		dims = (n,)
		dimnames = ('time',)
	cbh = np.zeros(dims, dtype=np.float64)
	for i in range(n):
		for j in range(l):
			x = cloud_mask[i,:,j] if len(dims) == 2 \
				else cloud_mask[i,:]
			kk = np.where(x)[0]
			if len(kk) > 0:
				if len(dims) == 2:
					cbh[i,j] = d['zfull'][kk[0]]
				else:
					cbh[i] = d['zfull'][kk[0]]
			else:
				if len(dims) == 2:
					cbh[i,j] = np.inf
				else:
					cbh[i] = np.inf
	d['cbh'] = cbh
	d['.']['cbh'] = {
		'.dims': dimnames,
		'long_name': 'cloud base height',
		'units': 'm',
	}

def stream(dd, state, **options):
	return misc.stream(dd, state, cloud_base_detection, **options)



import copy
import numpy as np
from alcf import misc
import ds_format as ds
import aquarius_time as aq

def tsample(d, tres):
	w = d['time_bnds'][:,1] - d['time_bnds'][:,0]
	d['time_bnds'] = np.array([[
		np.amin(d['time_bnds'][:,0]),
		np.amax(d['time_bnds'][:,1])
	]])
	d['time'] = np.array(np.mean(d['time_bnds'], axis=1))
	if 'backscatter_sd' in d:
		n = d['backscatter_sd'].shape[0]
		shape1 = list(d['backscatter_sd'].shape[1:])
		d['backscatter_sd'] = np.sqrt(1./n*np.average(
			d['backscatter_sd']**2,
			axis=0,
			weights=w,
		))
		d['backscatter_sd'] = d['backscatter_sd'].reshape([1] + shape1)
	for var in ds.get_vars(d):
		if var in ('time', 'time_bnds', 'backscatter_sd'):
			continue
		if 'time' not in d['.'][var]['.dims']:
			continue
		i = d['.'][var]['.dims'].index('time')
		shape = list(d[var].shape)
		d[var] = np.average(d[var], axis=i, weights=w)
		shape[i] = 1
		d[var] = d[var].reshape(shape)

def stream(dd, state, tres=None, align=True, **options):
	if tres is not None:
		state['aggregate_state'] = state.get('aggregate_state', {})
		dd = misc.aggregate(dd, state['aggregate_state'], tres, align=align)
		return misc.stream(dd, state, tsample, tres=tres)
	return dd



import sys
import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc
import aquarius_time as aq

VARIABLES = [
	'air_pressure',
	'air_temperature',
	'mass_fraction_of_cloud_liquid_water_in_air',
	'mass_fraction_of_cloud_ice_in_air',
	'cloud_volume_fraction_in_atmosphere_layer',
]

TRANS = {
	'air_pressure': 'pfull',
	'air_temperature': 'ta',
	'mass_fraction_of_cloud_liquid_water_in_air': 'clw',
	'mass_fraction_of_cloud_ice_in_air': 'cli',
	'cloud_volume_fraction_in_atmosphere_layer': 'cl',
}

STEP = 6/24

def read(dirname, index, track, t1, t2, warnings=[], step=STEP):
	dd_index = ds.readdir(dirname,
		variables=['time', 'latitude', 'longitude', 'level_height'],
		jd=True,
		full=True,
		warnings=warnings,
	)
	d_var = {}
	for var in VARIABLES:
		dd = []
		for d_index in dd_index:
			if var not in d_index['.']:
				continue
			time = d_index['time']
			time_half = misc.half(time)
			lat = d_index['latitude']
			lon = d_index['longitude']
			level_height = d_index['level_height']
			filename = d_index['filename']
			ii = np.nonzero((time_half[1:] >= t1) & (time_half[:-1] < t2))[0]
			for i in ii:
				t = time[i]
				lon0, lat0 = track(time[i])
				if np.isnan(lon0) or np.isnan(lat0):
					continue
				j = np.argmin(np.abs(lat - lat0))
				k = np.argmin(np.abs(lon - lon0))
				d = ds.read(filename, variables=[var],
					sel={'time': i, 'latitude': j, 'longitude': k})
				d_new = {
					'lat': np.array([lat[j]]),
					'lon': np.array([lon[k]]),
					'time': np.array([t]),
					'level_height': np.array([level_height]),
					'.': META,
				}
				d_new['.']['level_height'] = {
					'.dims': ['level']
				}
				d_new[TRANS[var]] = d[var].reshape([1] + list(d[var].shape))
				if TRANS[var] == 'cl':
					d_new[TRANS[var]] *= 100.
				dd.append(d_new)
		if len(dd) > 0:
			d_var[TRANS[var]] = ds.op.merge(dd, 'time')
	time_list = [
		set(d_var[var]['time'])
		for var in d_var.keys()
	]
	time = time_list[0].intersection(*time_list[1:]) \
		if len(time_list) > 0 \
		else set()
	d = {}
	d['.'] = {}
	if len(time) == 0:
		return None
	for var in d_var.keys():
		idx = [i for i, t in enumerate(d_var[var]['time']) if t in time]
		ds.select(d_var[var], {'time': idx})
		d[var] = d_var[var][var]
		d['lat'] = d_var[var]['lat']
		d['lon'] = d_var[var]['lon']
		d['time'] = d_var[var]['time']
		d['level_height'] = d_var[var]['level_height']
	if 'ps' not in d:
		n, m = d['pfull'].shape
		d['ps'] = np.full(n, np.nan, dtype=np.float64)
		for i in range(n):
	 		d['ps'][i] = 2*d['pfull'][i,0] - d['pfull'][i,1]
	if 'zfull' not in d:
		n, m = d['pfull'].shape
		d['zfull'] = np.full((n, m), np.nan, dtype=np.float64)
		for i in range(n):
			d['zfull'][i,:] = d['level_height']
	del d['level_height']
	if 'orog' not in d:
		n, m = d['zfull'].shape
		d['orog'] = np.full(n, np.nan, dtype=np.float64)
		for i in range(n):
			d['orog'][i] = 2*d['zfull'][i,0] - d['zfull'][i,1]
	return d



import copy
import numpy as np
import astropy.coordinates
import astropy.time
import astropy.units
import ds_format as ds
import aquarius_time as aq

def parse_time(time):
	if len(time) != 2:
		raise ValueError('Invalid time: %s' % time)
	start = aq.from_iso(time[0])
	end = aq.from_iso(time[1])
	if start is None or end is None:
		raise ValueError('Invalid time: %s' % time)
	return [start, end]

def aggregate(dd, state, period, epsilon=1./86400., align=True):
	dd = state.get('dd', []) + dd
	state['dd'] = []

	if len(dd) == 0 or dd[0] is None:
		return dd

	def merge(dd, t1, t2):
		if len(ddb) > 0:
			dx = ds.merge(ddb, 'time')
			dx['time_bnds'][0,0] = max(t1, dx['time_bnds'][0,0])
			dx['time_bnds'][-1,1] = min(t2, dx['time_bnds'][-1,1])
			if dx['time_bnds'][-1,1] > dx['time_bnds'][0,0]:
				return [dx]
		return []

	ddo = []
	ddb = []
	t = dd[0]['time_bnds'][0,0]
	if align:
		r = (t + 0.5 + epsilon) % period
		t1 = state.get('t1', t - r + epsilon)
	else:
		t1 = state.get('t1', t)
	t2 = state.get('t2', t1 + period)
	for d in dd:
		if d is None:
			ddo += merge(ddb, t1, t2) + [None]
			break
		i1 = 0
		i = 0
		while i < len(d['time']):
			if not (d['time_bnds'][i,1] - t2 > epsilon):
				i += 1
				continue
			ii = np.arange(i1, i + (t2 - d['time_bnds'][i,0] > epsilon))
			if len(ii) > 0:
				dx = copy.copy(d)
				ds.select(dx, {'time': ii})
				ddb += [dx]
			ddo += merge(ddb, t1, t2)
			i1 = i
			t = d['time_bnds'][i,0]
			if align:
				r = (t + 0.5 + epsilon) % period
				t1 = max(t2, t - r + epsilon)
			else:
				t1 = max(t2, t)
			t2 = t1 + period
			ddb = []
		ii = np.arange(i1, len(d['time']))
		if len(ii) > 0:
			dx = copy.copy(d)
			ds.select(dx, {'time': ii})
			ddb += [dx]
	state['dd'] = ddb
	state['t1'] = t1
	state['t2'] = t2
	return ddo

def stream(dd, state, f, **options):
	dd = dd + state.get('dd', [])
	i = 0
	for i, d in enumerate(dd):
		if d is None:
			break
		f(d, **options)
	state['dd'] = []
	return dd[:(i+1)]

def half(xfull):
	xhalf = np.zeros(len(xfull) + 1, dtype=xfull.dtype)
	xhalf[1:-1] = 0.5*(xfull[1:] + xfull[:-1])
	xhalf[0] = 2.*xfull[0] - xfull[1]
	xhalf[-1] = 2.*xfull[-1] - xfull[-2]
	return xhalf

def time_bnds(time, step, start=None, end=None):
	n = len(time)
	bnds = np.full((n, 2), np.nan, time.dtype)
	bnds[:,0] = time - step*0.5
	bnds[:,1] = time + step*0.5
	bnds[1:,0] = np.maximum(bnds[:-1,1], bnds[1:,0])
	bnds[:-1,1] = np.minimum(bnds[1:,0], bnds[:-1,1])
	if start is not None:
		bnds[0,0] = max(bnds[0,0], start)
	if end is not None:
		bnds[-1,1] = min(bnds[-1,1], end)
	return bnds

def sun_altitude(t, lon, lat):
	loc = astropy.coordinates.EarthLocation(
		lon=lon*astropy.units.deg,
		lat=lat*astropy.units.deg
	)
	time = astropy.time.Time(t, format='jd')
	altaz = astropy.coordinates.AltAz(location=loc, obstime=time)
	sun = astropy.coordinates.get_sun(time)
	return sun.transform_to(altaz).alt.hour/24.*360.

def require_vars(d, variables):
	for v in variables:
		if v not in d:
			raise ValueError('Variable "%s" is required' % v)

def geo_distance(lon1, lat1, lon2, lat2, method='gc'):
	lon1, lat1, lon2, lat2 = [x/180*np.pi for x in (lon1, lat1, lon2, lat2)]
	if method == 'gc':
		x = np.sin(lat1)*np.sin(lat2) + \
			np.cos(lat1)*np.cos(lat2)*np.cos(lon1 - lon2)
		return 6371*(np.arccos(np.maximum(np.minimum(x, 1), -1)))
	elif method == 'hs':
		dlon = lon2 - lon1
		dlat = lat2 - lat1
		a = np.sin(dlat/2)**2 + np.cos(lat1)*np.cos(lat2)*np.sin(dlon/2)**2
		return 2*6371*np.arcsin(np.sqrt(a))
	else:
		raise ValueError('Unrecognized method "%s"' % method)

def time_mask(bnds, t1, t2):
	return ~(((bnds[:,0] < t1) & (bnds[:,1] < t1)) |
	         ((bnds[:,0] > t2) & (bnds[:,1] > t2)))

def track_at(d, t, epsilon=1/86400):
	i = np.searchsorted(d['time_bnds'][:,0], t, side='right')
	if i == 0:
		if t >= d['time_bnds'][0,0] - epsilon:
			return d['lon'][0], d['lat'][0]
		else:
			return np.nan, np.nan
	if t <= d['time_bnds'][i-1,1] + epsilon:
		return d['lon'][i-1], d['lat'][i-1]
	return np.nan, np.nan

def populate_meta(d, meta, vars):
	d_tmp = {'.': meta}
	for var in vars:
		ds.meta(d, var, ds.meta(d_tmp, var))
	ds.meta(d, None, ds.meta(d_tmp))

def keep_var(var, d, do, dim_map={}):
	dim_map_rev = dict((v, k) for k, v in dim_map.items())
	if var in ds.vars(d) and dim_map.get('time', 'time') in ds.dims(d, var):
		name = 'input_' + var
		ds.var(do, name, ds.var(d, var).astype(np.float64))
		dims = [dim_map_rev.get(dim, dim) for dim in ds.dims(d, var)]
		ds.dims(do, name, dims)
		ds.attrs(do, name, ds.attrs(d, var))
		ds.rm_attr(do, '_FillValue', name)

def dep_vars(mapping, vars):
	return list(set([y for x in vars if x in mapping for y in mapping[x]]))

def point_to_track(point, time):
	time_mid = 0.5*(time[0] + time[1])
	return {
		'lon': np.array([point[0] % 360, point[0] % 360], dtype=np.float64),
		'lat': np.array([point[1], point[1]], dtype=np.float64),
		'time': np.array([time[0], time[1]], dtype=np.float64),
		'time_bnds': np.array([[time[0], time_mid], [time_mid, time[1]]],
			dtype=np.float64),
	}

def track_auto_time_bnds(time, track_gap=0):
	n = len(time)
	time_bnds = np.full((n, 2), np.nan, np.float64)
	time_bnds[0,0] = time[0]
	time_bnds[-1,1] = time[-1]
	time_avg = 0.5*(time[:-1] + time[1:])
	time_bnds[1:,0] = time_avg
	time_bnds[:-1,1] = time_avg
	if track_gap != 0:
		time_diff = time[1:] - time[:-1]
		mask1 = np.full(n, False, bool)
		mask2 = np.full(n, False, bool)
		mask1[:-1] = time_diff > track_gap
		mask2[1:] = time_diff > track_gap
		time_bnds[mask1,1] = time[mask1]
		time_bnds[mask2,0] = time[mask2]
	return time_bnds

def read_track(filenames, lon_180=False, track_gap=0):
	if type(filenames) not in [list, tuple]:
		filenames = [filenames]
	dd = []
	for filename in filenames:
		d = ds.read(filename, jd=True)
		if len(d['time']) < 2:
			raise ValueError('%s: Track must contain at least two records', filename)
		if 'time_bnds' not in d:
			d['time_bnds'] = track_auto_time_bnds(d['time'], track_gap)
			d['.']['time_bnds'] = {
				'.dims': ['time', 'bnds'],
				'long_name': 'time bounds',
				'standard_name': 'time',
				'units': 'days since -4713-11-24 12:00 UTC',
				'calendar': 'proleptic_gregorian',
			}
		dd += [d]
	d = ds.merge(dd, 'time')
	d['lon'] = d['lon'] % 360
	return d

def track_has_seg(track, t1, t2):
	mask = (track['time_bnds'][:,0] < t2) & (track['time_bnds'][:,1] >= t1)
	return np.any(mask)

def cmd_point_or_track(point, time, track, track_gap=0):
	time_lim = [-np.inf, np.inf]
	if time is not None:
		for i in [0, 1]:
			time_lim[i] = aq.from_iso(time[i])
			if time_lim[i] is None:
				raise ValueError('Invalid time format: %s' % time[i])
	d = None
	if track is not None:
		d = read_track(track, track_gap/86400.)
	elif point is not None and time is not None:
		d = point_to_track(point, time_lim)
	else:
		raise ValueError('Point and time or track is required')
	return d, time_lim



import os
from getpass import getpass
import requests
	
PRODUCTS = ['M2I3NVASM']

URS = b'urs.earthdata.nasa.gov'

TEMPLATE = 'https://goldsmr5.gesdisc.eosdis.nasa.gov/daac-bin/OTF/HTTP_services.cgi?FILENAME=%2Fdata%2FMERRA2%2FM2I3NVASM.5.12.4%2F{year:04d}%2F{month:02d}%2FMERRA2_{gen}.inst3_3d_asm_Nv.{year:04d}{month:02d}{day:02d}.nc4&LABEL=MERRA2_{gen}.inst3_3d_asm_Nv.{year:04d}{month:02d}{day:02d}.SUB.nc&SERVICE=L34RS_MERRA2&DATASET_VERSION=5.12.4&FORMAT=bmM0Lw&VERSION=1.02&SHORTNAME=M2I3NVASM&BBOX={lat1:.3f}%2C{lon1:.3f}%2C{lat2:.3f}%2C{lon2:.3f}&VARIABLES=CLOUD%2CH%2CPHIS%2CPL%2CQI%2CQL%2CPS%2CT'

def quote(s):
	return s.replace(b'\\', b'\\\\').replace(b'"', b'\\"')

def login(user=None, password=None, overwrite=False):
	if user is None:
		user = input('NASA Earthdata login: ')
	if password is None:
		password = getpass(prompt='NASA Earthdata password: ')
	if type(user) is not str:
		user = str(user)
	if type(password) is not str:
		password = str(password)

	home = os.path.expanduser("~")
	netrc = os.path.join(home, '.netrc')
	urs_cookies = os.path.join(home, '.urs_cookies')
	if os.name == 'nt':
		dodsrc = os.path.join(os.getcwd(), '.dodsrc')
	else:
		dodsrc = os.path.join(home, '.dodsrc')

	ask_files = [x for x in [netrc, urs_cookies, dodsrc] if os.path.exists(x)]
	if len(ask_files) > 0 and overwrite is False:
		print('The following files have to be overwritten:')
		for file_ in ask_files:
			print('    ' + file_)
		res = input('Continue [y/n]? ')
		if res != 'y':
			return

	with open(netrc, 'wb', opener=lambda p, f: os.open(p, f, mode=0o600)) as f:
		os.chmod(f.fileno(), 0o600)
		f.write(b'machine %s login "%s" password "%s"\n' % (
			URS,
			quote(user.encode('utf-8')),
			quote(password.encode('utf-8'))
		))
	print('-> %s' % netrc)

	with open(urs_cookies, 'wb') as f:
		pass
	print('-> %s' % urs_cookies)

	with open(dodsrc, 'wb') as f:
		f.write(b'HTTP.COOKIEJAR=%s\n' % os.fsencode(urs_cookies))
		f.write(b'HTTP.NETRC=%s\n' % os.fsencode(netrc))
	print('-> %s' % dodsrc)

def download(output, product, year, month, day, lon1, lon2, lat1, lat2,
	nocache=False
):
	lon1_180 = lon1 if lon1 < 180 else lon1 - 360
	lon2_180 = lon2 if lon2 < 180 else lon2 - 360
	url = TEMPLATE.format(
		year=year, month=month, day=day,
		lat1=lat1, lat2=lat2,
		lon1=lon1_180, lon2=lon2_180,
		gen=(300 if year < 2011 else 400),
	)
	response = requests.get(url)
	with open(output, 'wb') as f:
		f.write(response.content)



import os
import numpy as np
import ds_format as ds
import aquarius_time as aq
from alcf.algorithms import tsample, zsample
from alcf import misc

VARIABLES = [
	'time',
	'cloud_mask',
	'zfull',
]

def read(d):
	print('<- %s' % d['filename'])
	if d is not None:
		d0 = ds.read(d['filename'], VARIABLES)
		d.update(d0)

def read_stream(dd, state):
	return misc.stream(dd, state, read)

def compare_stream(ddd, state):
	n = len(ddd)
	state['ddd'] = state.get('ddd', [[] for i in range(n)])
	for i in range(n):
		state['ddd'][i] += ddd[i]
	if np.all(np.array([
		len(state['ddd'][i]) > 0
		for i in range(n)
	])):
		for i in range(n):
			time = state['ddd'][i][0]['time']
			t1 = time[0]
	return []

def stream(ddd, state):
	n = len(ddd)
	state['ddd'] = state.get('ddd', [[] for i in range(n)])
	state['read_stream_state'] = state.get('read_stream_state', {})
	state['compare_stream_state'] = state.get('compare_stream_state', {})
	compare_stream([[] for i in range(n)], state['compare_stream_state'])
	for i, dd in enumerate(ddd):
		state['ddd'][i] += dd
	while True:
		for i in range(n):
			if len(state['compare_stream_state']['ddd'][i]) == 0 and \
				len(state['ddd'][i]) > 0:
				dd = [state['ddd'][i][0]]
				state['ddd'][i] = state['ddd'][i][1:]
				dd = read_stream(dd, state['read_stream_state'])
				ddd0 = [[] for j in range(n)]
				ddd0[i] = dd
				dd = compare_stream(ddd0, state['compare_stream_state'])
	return dd

def run(arg1, arg2, arg3, *args):
	'''
alcf-compare -- Calculate comparison statistics from multiple lidar time series.
============

Synopsis
--------

    alcf compare <input_1> <input_2> [<input_n>...] <output>

Arguments
---------

- `input_*`: Input lidar data directory.
- `output`: Output filename.
	'''
	args = [arg1, arg2, arg3] + list(args)
	input_ = args[:-1]
	output = args[-1]

	n = len(input_)
	files = [
		[
			{'filename': os.path.join(dirname, f)}
			for f in sorted(os.listdir(dirname))
			if os.path.isfile(os.path.join(dirname, f))
		]
		for dirname in input_
	]
	state = {}
	stream(files, state)
	dd = stream([[None] for i in range(n)], state)
	print('-> %s' % output)
	#ds.to_netcdf(output, dd[0])



import ds_format as ds
import os
import re
import numpy as np
import aquarius_time as aq
from alcf.models import META
from alcf import misc

KAPPA = 0.2854 # Poisson constant for dry air.

# Variables in NetCDF files produces by the AMPS project (probably discontinued).
VARS_NC = [
	'QCLOUD',
	'QICE',
	'PSFC',
	'P',
	'PB',
	'PHB',
	'PH',
	'HGT',
	'XTIME',
	'XLONG',
	'XLAT',
	'T',
]

# Variables in NetCDF files converted from GRIB with ncl_convert2nc.
VARS_NCL = [
	'g5_lon_1',
	'g5_lat_0',
	'PRES_GDS5_SFC',
	'HGT_GDS5_SFC',
	'HGT_GDS5_ISBL',
	'TMP_GDS5_ISBL',
	'CLWMR_GDS5_ISBL',
	'ICMR_GDS5_ISBL',
]

STEP = 3/24

def index(dirname, warnings=[], recursive=False, njobs=1):
	return ds.readdir(dirname, ['XTIME'],
		jd=True,
		recursive=recursive,
		warnings=warnings,
		parallel=(njobs > 1),
		njobs=njobs,
		full=True,
	)

def read_nc(d_index, track, t1, t2, step):
	time = d_index['XTIME'][0]
	time0 = d_index['.']['.']['SIMULATION_START_DATE']
	time0 = aq.from_iso(time0.replace('_', 'T'))
	if time < 2000000.:
		time = time0 + time/(24*60)
	if (time >= t1 - step*0.5) & (time < t2 + step*0.5):
		lon0, lat0 = track(time)
		if np.isnan(lon0) or np.isnan(lat0):
			return
		d = ds.read(d_index['filename'], variables=VARS_NC, sel={'Time': 0})
		lon = np.where(d['XLONG'] < 0, 360 + d['XLONG'], d['XLONG'])
		lat = d['XLAT']
		l = np.argmin((lon - lon0)**2 + (lat - lat0)**2)
		i, j = np.unravel_index(l, lon.shape)
		clw = d['QCLOUD'][:,i,j]
		cli = d['QICE'][:,i,j]
		cl = np.full(len(clw), 100, dtype=np.float64)
		ps = d['PSFC'][i,j]
		orog = d['HGT'][i,j]
		pfull = d['PB'][:,i,j] + d['P'][:,i,j]
		zfull = (d['PHB'][:,i,j] + d['PH'][:,i,j])/9.81
		zfull = 0.5*(zfull[1:] + zfull[:-1])
		theta = d['T'][:,i,j] + 300
		ta = theta*(pfull/ps)**KAPPA
		newshape3 = [1] + list(clw.shape)
		newshape2 = [1] + list(ps.shape)
		return {
			'clw': clw.reshape(newshape3),
			'cli': cli.reshape(newshape3),
			'ta': ta.reshape(newshape3),
			'cl': cl.reshape(newshape3),
			'pfull': pfull.reshape(newshape3),
			'zfull': zfull.reshape(newshape3),
			'ps': ps.reshape(newshape2),
			'orog': orog.reshape(newshape2),
			'lon': np.array([lon[i,j]]),
			'lat': np.array([lat[i,j]]),
			'time': np.array([time]),
			'.': META,
		}

def read_ncl(d_index, track, t1, t2, step):
	s = d_index['.']['.']['grib_source']
	m = re.match(r'^(\d\d\d\d)(\d\d)(\d\d)(\d\d)_.*_f(\d\d\d)\.grb', s)
	g = m.groups()
	year = int(g[0])
	month = int(g[1])
	day = int(g[2])
	hour = int(g[3])
	forecast = int(g[4])
	time = aq.from_date([1, year, month, day, hour]) + forecast/24

	dim1 = ds.dims(d_index, 'HGT_GDS5_ISBL')[0]
	dim2 = ds.dims(d_index, 'CLWMR_GDS5_ISBL')[0]

	if (time >= t1 - step*0.5) & (time < t2 + step*0.5):
		lon0, lat0 = track(time)
		if np.isnan(lon0) or np.isnan(lat0):
			return
		d = ds.read(d_index['filename'], variables=VARS_NCL + [dim1, dim2])
		lon = np.where(d['g5_lon_1'] < 0, 360 + d['g5_lon_1'], d['g5_lon_1'])
		lat = d['g5_lat_0']
		l = np.argmin((lon - lon0)**2 + (lat - lat0)**2)
		i, j = np.unravel_index(l, lon.shape)
		k1 = ds.dim(d, dim1)
		k2 = ds.dim(d, dim2)
		dk = k1 - k2
		clw = d['CLWMR_GDS5_ISBL'][::-1,i,j]
		cli = d['ICMR_GDS5_ISBL'][::-1,i,j]
		cl = np.full(len(clw), 100, dtype=np.float64)
		ps = d['PRES_GDS5_SFC'][i,j]
		orog = d['HGT_GDS5_SFC'][i,j]
		pfull = d[dim2][::-1]*100
		zfull = d['HGT_GDS5_ISBL'][dk:,i,j][::-1]
		ta = d['TMP_GDS5_ISBL'][dk:,i,j][::-1]
		newshape3 = [1] + list(clw.shape)
		newshape2 = [1] + list(ps.shape)
		return {
			'clw': clw.reshape(newshape3),
			'cli': cli.reshape(newshape3),
			'ta': ta.reshape(newshape3),
			'cl': cl.reshape(newshape3),
			'pfull': pfull.reshape(newshape3),
			'zfull': zfull.reshape(newshape3),
			'ps': ps.reshape(newshape2),
			'orog': orog.reshape(newshape2),
			'lon': np.array([lon[i,j]]),
			'lat': np.array([lat[i,j]]),
			'time': np.array([time]),
			'.': META,
		}

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	dd = []
	for d_index in index:
		if 'XTIME' in d_index:
			d = read_nc(d_index, track, t1, t2, step)
		else:
			d = read_ncl(d_index, track, t1, t2, step)
		if d is not None:
			dd.append(d)
	d = ds.op.merge(dd, 'time')
	return d



import numpy as np
import ds_format as ds
import datetime as dt
from alcf import misc
from alcf.lidars import META

WAVELENGTH = 532 # nm
CALIBRATION_COEFF = 3.75e-6
SURFACE_LIDAR = True
SC_LR = 16.0 # sr
MAX_RANGE = 30000 # m

VARS = {
	'backscatter': ['nrb_copol', 'nrb_crosspol'],
	'zfull': ['bin_time', 'c'],
}

DEFAULT_VARS = [
	'time',
	'elevation_angle',
	'gps_altitude',
	'gps_latitude',
	'gps_longitude',
]

def read(
	filename,
	vars,
	altitude=None,
	lon=None,
	lat=None,
	tlim=None,
	keep_vars=[],
	**kwargs
):
	sel = None
	tres = None
	if tlim is not None:
		d = ds.read(filename, 'time', jd=True)
		tres = d['time'][1] - d['time'][0]
		d['time_bnds'] = misc.time_bnds(d['time'], tres)
		mask = misc.time_mask(d['time_bnds'], tlim[0], tlim[1])
		if np.sum(mask) == 0: return None
		sel = {'profile': mask}

	dep_vars = misc.dep_vars(VARS, vars)
	req_vars = dep_vars + DEFAULT_VARS + keep_vars
	d = ds.read(filename, req_vars, jd=True, sel=sel, full=True)
	mask = d['elevation_angle'] == 0.0
	dx = {}
	misc.populate_meta(dx, META, set(vars) & set(VARS))
	n = ds.dim(d, 'profile')
	m = ds.dim(d, 'range')
	altitude = d['gps_altitude'] if altitude is None else \
		np.full(n, altitude, np.float64)
	lon = d['gps_longitude'] if lon is None else \
		np.full(n, lon, np.float64)
	lat = d['gps_latitude'] if lat is None else \
		np.full(n, lat, np.float64)
	if 'time' in vars:
		dx['time'] = d['time']
	if 'time_bnds' in vars:
		if tres is None: tres = d['time'][1] - d['time'][0]
		args = [] if tlim is None else [tlim[0], tlim[1]]
		dx['time_bnds'] = misc.time_bnds(d['time'], tres, *args)
	if 'zfull' in vars:
		dx['zfull'] = np.full((n, m), np.nan, np.float64)
		for i in range(n):
			range_ = 0.5*d['bin_time'][i]*d['c']*(np.arange(m) + 0.5)
			dx['zfull'][i,:] = range_*np.sin(d['elevation_angle'][i]/180.0*np.pi)
			dx['zfull'][i,:] += altitude[i]
	if 'backscatter' in vars:
		dx['backscatter'] = (d['nrb_copol'] + 2.*d['nrb_crosspol'])*CALIBRATION_COEFF
	if 'altitude' in vars:
		dx['altitude'] = altitude
	if 'lon' in vars:
		dx['lon'] = lon
	if 'lat' in vars:
		dx['lat'] = lat
	for var in keep_vars:
		misc.keep_var(var, d, dx, {'time': 'profile', 'level': 'range'})
	return dx



import numpy as np
import ds_format as ds
import datetime as dt
from alcf import misc
from alcf.lidars import META

WAVELENGTH = 910 # nm
CALIBRATION_COEFF = 1.2e-3
SURFACE_LIDAR = True
SC_LR = 18.8 # sr. Stratocumulus lidar ratio (O'Connor et al., 2004).
MAX_RANGE = 15400 # m

VARS = {
	'backscatter': ['backscatter'],
}

DEFAULT_VARS = [
	'vertical_resolution',
	'level',
	'range',
	'time',
	'detection_status',
]

def read(filename, vars,
	altitude=None,
	lon=None,
	lat=None,
	calibration_coeff=CALIBRATION_COEFF,
	fix_cl_range=False,
	cl_crit_range=6000,
	tlim=None,
	keep_vars=[],
	**kwargs
):
	sel = None
	tres = None
	if tlim is not None:
		d = ds.read(filename, 'time', jd=True)
		tres = d['time'][1] - d['time'][0]
		d['time_bnds'] = misc.time_bnds(d['time'], tres)
		mask = misc.time_mask(d['time_bnds'], tlim[0], tlim[1])
		if np.sum(mask) == 0: return None
		sel = {'time': mask}

	dep_vars = misc.dep_vars(VARS, vars)
	req_vars = dep_vars + DEFAULT_VARS + keep_vars
	d = ds.read(filename, req_vars, jd=True, sel=sel, full=True)
	dx = {}
	misc.populate_meta(dx, META, set(vars) & set(VARS))
	n = ds.dim(d, 'time')
	if 'time' in vars:
		dx['time'] = d['time']
	if 'time_bnds' in vars:
		if tres is None: tres = d['time'][1] - d['time'][0]
		args = [] if tlim is None else [tlim[0], tlim[1]]
		dx['time_bnds'] = misc.time_bnds(d['time'], tres, *args)
	if 'range' in d: # ARM CL51 format.
		range_ = d['range']
	else: # Generic CL51 format.
		range_ = d['vertical_resolution'][0]*d['level']
	if 'zfull' in vars:
		zfull1 = range_
		dx['zfull'] = np.tile(zfull1, (n, 1))
		if altitude is not None:
			dx['zfull'] += altitude
	if 'backscatter' in vars:
		factor = 1e-4 if (d['.']['backscatter']['units'] == '1/(sr*km*10000)') \
			else 1 # Factor of 1e-4 if ARM CL51 format.
		dx['backscatter'] = d['backscatter']*calibration_coeff*factor
		mask = range_ > 6000
		if fix_cl_range is True:
			for i in range(n):
				if d['detection_status'][i] == b'0':
					dx['backscatter'][i,mask] *= (range_[mask]/6000)**2
	if 'altitude' in vars:
		dx['altitude'] = np.full(n, altitude, np.float64)
	if 'lon' in vars:
		dx['lon'] = np.full(n, lon, np.float64)
	if 'lat' in vars:
		dx['lat'] = np.full(n, lat, np.float64)
	for var in keep_vars:
		misc.keep_var(var, d, dx)
	return dx



import numpy as np
import ds_format as ds
from alcf import misc
from alcf.algorithms import interp

def zsample(d, zres=None, zlim=None):
	n = ds.dim(d, 'time')
	m = ds.dim(d, 'level')
	l = ds.dim(d, 'column')
	if m == 0:
		return

	zfull = d['zfull']
	zhalf = np.zeros((n, m+1), dtype=np.float64)
	for i in range(n):
		zhalf[i,:] = misc.half(zfull[i,:]) \
			if zfull.ndim == 2 \
			else misc.half(zfull)
	zhalf2 = np.arange(zlim[0], zlim[-1] + zres, zres, np.float64)
	zfull2 = (zhalf2[1:] + zhalf2[:-1])*0.5
	m2 = len(zfull2)

	for var in ds.vars(d):
		dims = ds.dims(d, var)
		if var == 'zfull' or 'level' not in dims:
			continue
		x = ds.var(d, var)
		i = dims.index('time')
		j = dims.index('level')
		x = np.moveaxis(x, [i, j], [0, 1])
		shape = list(x.shape)
		x = x.reshape(n, m, -1)
		shape2 = list(x.shape)
		shape2[1] = m2
		l = shape2[2]
		x2 = np.zeros(shape2, dtype=x.dtype)
		if var == 'backscatter_sd':
			x = x**2
		for i2 in range(n):
			for j2 in range(l):
				x2[i2,:,j2] = interp(zhalf[i2,:], x[i2,:,j2], zhalf2)
		if var == 'backscatter_sd':
			x2 = np.sqrt(x2)
		x2 = x2.reshape([n, m2] + shape[2:])
		x2 = np.moveaxis(x2, [0, 1], [i, j])
		ds.var(d, var, x2)

	ds.var(d, 'zfull', zfull2)
	ds.dims(d, 'zfull', ['level'])

def stream(dd, state, zres=None, zlim=None, **options):
	return misc.stream(dd, state, zsample, zres=zres, zlim=zlim)



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

# (c) British Crown Copyright, the Met Office.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted 
# provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice, this list 
#       of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice, this list
#       of conditions and the following disclaimer in the documentation and/or other materials 
#       provided with the distribution.
#     * Neither the name of the Met Office nor the names of its contributors may be used 
#       to endorse or promote products derived from this software without specific prior written 
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER 
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import matplotlib.cm as mpl_cm
import matplotlib.pyplot as plt
import iris
import iris.quickplot as qplt
import numpy as np
import iris.analysis.cartography as icart

dbze94 = 'equivalent_reflectivity_factor'
atb532 = 'volume_attenuated_backwards_scattering_function_in_air'
cfadDbze94 = 'histogram_of_equivalent_reflectivity_factor_over_height_above_reference_ellipsoid'
cfadLidarsr532 = 'histogram_of_backscattering_ratio_over_height_above_reference_ellipsoid'
clisccp = 'cloud_area_fraction_in_atmosphere_layer'


######### area_average ########################################################
def area_average(x,latitude='latitude',longitude='longitude'):

    if x.coord(latitude).bounds is None:
        x.coord(latitude).guess_bounds()
    if x.coord(longitude).bounds is None:
        x.coord(longitude).guess_bounds()
    grid_areas = icart.area_weights(x)
    x_avg = x.collapsed([longitude, latitude], iris.analysis.MEAN, 
                            weights=grid_areas)
    return x_avg

######### change_data_units ##################################################
def change_data_units(x,std_name,new_units='%',factor=100):
    """
    Function that changes the units of a cube.
    """

    x = x * factor
    x.standard_name = std_name
    x.units = new_units
    return x

######### change_coord_units #################################################
def change_coord_units(x,coord_name='height',new_units='km',factor=0.001):
    """
    Function that changes the units of a coordinate.
    """

    coord = x.coord(coord_name)
    coord.points = coord.points * factor
    coord.units = new_units
    if coord.bounds is not None:
        coord.bounds = coord.bounds * factor

############### change_pseudo_level ###########################################
def change_pseudo_level(x, origin='pp'):
    """
    Function that changes the pp pseudo_level dimension to the correct units.
    """
    pname = 'pseudo_level'
    # Calculate values
    if x.standard_name == cfadDbze94:
        deltas = np.arange(15)*5
        points = -47.5 + deltas
        bounds = np.array( [ (-50.0 + deltas), (-45.0 + deltas) ] )
        bounds = bounds.transpose()
        units  = 'dBZ'
        std_name = 'equivalent_reflectivity_factor'
    elif x.standard_name == cfadLidarsr532:
        #points = np.array([0.005, 0.605, 2.1, 4, 6, 8.5, 12.5, 17.5, 
                           #22.5, 27.5, 35, 45, 55, 70, 50040])
        #vb = np.array([0, 0.01, 1.2, 3, 5, 7, 10, 15, 20, 
                      #25, 30, 40, 50, 60, 80, 100000])
        #bounds =   np.empty( (15,2), dtype=float)
        #bounds[:,0] = vb[0:15]
        #bounds[:,1] = vb[1:16]
        deltas = np.arange(15)
        points = 0.5 + deltas
        bounds = np.array( [ (0 + deltas), (1 + deltas) ] )
        bounds = bounds.transpose()
        units = '1'
        std_name = 'backscattering_ratio'
        if origin == 'nc':
            pname = std_name
    elif x.standard_name == clisccp:
        deltas = np.arange(7)
        points = 0.5 + deltas
        bounds = np.array( [ (0 + deltas), (1 + deltas) ] )
        bounds = bounds.transpose()
        units = '1'
        std_name = 'atmosphere_optical_thickness_due_to_cloud'
        if origin == 'nc':
            pname = std_name
            points = points[1:]
            bounds = bounds[1:][:]
    else:
        raise NameError('Pseudo level not recognised: ' + x.standard_name)

    # Assign values
    pseudo = x.coord(pname)
    pseudo.points = points
    pseudo.units = units
    pseudo.standard_name = std_name
    pseudo.bounds = bounds

######### cosp_plot_column_1D ##########################################################
def cosp_plot_column_1D(fnc, varname='equivalent_reflectivity_factor', column = 0, time = 0):
    """
    Function that plots one column of curtain data.
    """

    plt.interactive(True)
    fig=plt.figure()
    # Read cube
    z=iris.load(fnc)
    z=z[0]

    # Get coords
    c = z.coord('column')
    t = z.coord('time')

    # Select time and column
    y=z.extract(iris.Constraint(column=c.points[column]))
    y=y.extract(iris.Constraint(time=t.points[time]))

    qplt.pcolormesh(y)

    return

######### cosp_plot_column_2D ##########################################################
def cosp_plot_column_2D(fnc, varname='equivalent_reflectivity_factor', level=0, column = 0, time = 0):
    """
    Function that plots one column of lat/lon data.
    """

    plt.interactive(True)
    fig=plt.figure()
    ax = fig.add_subplot(111)
    # Read cube
    z=iris.load(fnc)
    z=z[0]

    # Get coords
    c = z.coord('column')
    t = z.coord('time')

    # Select time and column
    y=z.extract(iris.Constraint(column=c.points[column]))
    y=y.extract(iris.Constraint(time=t.points[time]))
    # Select level. Not managed to make constrain with 'atmospheric model level'
    y=y[level]

    color_map = mpl_cm.get_cmap('Paired')
    qplt.pcolormesh(y,cmap=color_map,vmin=-20,vmax=20)
    plt.gca().coastlines()

    return


######### cosp_plot_cfad_2D ##########################################################
def cosp_plot_cfad_2D(fnc, varname=cfadDbze94, time = 0):
    """
    Function that plots an area-averaged CFAD from a 2D file.
    """

    # Read cube
    z=iris.load(fnc)
    z=z[0]

    # Select time
    t = z.coord('time')
    z=z.extract(iris.Constraint(time=t.points[time]))


    # Obtain area average
    cfad = area_average(z)

    if varname is not clisccp:
        cfad = change_data_units(cfad,varname)
        change_coord_units(cfad,coord_name='altitude')

    if cfad.standard_name in [cfadLidarsr532, clisccp]:
        change_pseudo_level(cfad, origin = 'nc')

    # Set limits
    yticks = None
    if cfad.standard_name == cfadDbze94:
        xlim = [-30, 20]
        ylim = [  0, 18]
        vmax = 8
    elif cfad.standard_name == cfadLidarsr532:
        xlim = [ 3, 15]
        ylim = [ 0, 18]
        vmax = 3
        xt1 = xlim[0]
        xt2 = xlim[1] + 1
        xticks  = xt1 + np.arange(xt2 - xt1)
        xlabels = ['0', '0.01', '1.2', '3', '5', '7', '10', '15', '20', 
                  '25', '30', '40', '50', '60', '80', '']
    elif cfad.standard_name == clisccp:
        xlim = [ 0, 7]
        ylim = [ 1000, 100]
        vmax = 8
        xt1 = xlim[0]
        xt2 = xlim[1] + 1
        xticks  = xt1 + np.arange(xt2 - xt1)
        xlabels = ['0', '0.3', '1.3', '3.6', '9.4', '23', '60', '']
        yticks = [1000, 800, 680, 560, 440, 310, 180]
    else:
        raise NameError('Variable not supported: ' + x.standard_name)

    # Plot block plot
    plt.interactive(True)
    fig=plt.figure()
    ax = fig.add_subplot(111)
    qplt.pcolormesh(cfad,vmin=0,vmax=vmax)
    # Title and limits
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    # Xlabels
    if cfad.standard_name == cfadLidarsr532:
        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels[xt1:xt2])
    if cfad.standard_name == clisccp:
        ax.set_xticklabels(xlabels)
    # Ylabels
    if yticks is not None:
        ax.set_yticks(yticks)

    return


import os
from alcf.cmds import lidar, stats, plot

STEPS = ['lidar', 'stats', 'plot']

def run(type_, input_, output, *args, skip=None, **kwargs):
	lidar_dir = os.path.join(output, 'lidar')
	stats_dir = os.path.join(output, 'stats')
	stats_filename = os.path.join(stats_dir, 'all.nc')
	stats_fine_filename = os.path.join(stats_dir, 'all_fine.nc')
	stats_clear_fine_filename = os.path.join(stats_dir, 'clear_fine.nc')
	plot_dir = os.path.join(output, 'plot')
	backscatter_dir = os.path.join(plot_dir, 'backscatter')
	cloud_occurrence_filename = os.path.join(plot_dir, 'cloud_occurrence.png')
	cbh_filename = os.path.join(plot_dir, 'cbh.png')
	backscatter_hist_filename = os.path.join(plot_dir, 'backscatter_hist.png')
	backscatter_hist_fine_filename = os.path.join(plot_dir, 'backscatter_hist_all_fine.png')
	backscatter_hist_clear_fine_filename = os.path.join(plot_dir, 'backscatter_hist_clear_fine.png')

	if skip is not None:
		try: i = STEPS.index(skip)
		except ValueError:
			raise ValueError('Invalid step "%s"' % skip)
	else:
		i = -1

	print('-> %s' % output)
	try: os.mkdir(output)
	except OSError: pass

	if i < STEPS.index('lidar'):
		print('-> %s' % lidar_dir)
		try: os.mkdir(lidar_dir)
		except OSError: pass
		print('! alcf lidar')
		lidar.run(type_, input_, lidar_dir, *args, **kwargs)
	if i < STEPS.index('stats'):
		print('-> %s' % stats_dir)
		try: os.mkdir(stats_dir)
		except OSError: pass
		print('! alcf stats')
		stats.run(lidar_dir, stats_filename, **kwargs)
		stats.run(lidar_dir, stats_fine_filename,
			blim=[-1, 1],
			bres=0.005,
			**kwargs
		)
		stats.run(lidar_dir, stats_clear_fine_filename,
			filter='clear',
			blim=[-1, 1],
			bres=0.005,
			**kwargs
		)
	if i < STEPS.index('plot'):
		print('-> %s' % plot_dir)
		try: os.mkdir(plot_dir)
		except OSError: pass
		print('-> %s' % backscatter_dir)
		try: os.mkdir(backscatter_dir)
		except OSError: pass
		print('! alcf plot backscatter')
		plot.run('backscatter', lidar_dir, backscatter_dir, **kwargs)
		print('! alcf plot cloud_occurrence')
		plot.run('cloud_occurrence', stats_filename, cloud_occurrence_filename,
			**kwargs
		)
		print('! alcf plot cbh')
		plot.run('cbh', stats_filename, cbh_filename,
			**kwargs
		)
		print('! alcf plot backscatter_hist')
		plot.run('backscatter_hist', stats_filename, backscatter_hist_filename,
			**kwargs
		)
		plot.run(
			'backscatter_hist',
			stats_fine_filename,
			backscatter_hist_fine_filename,
			vlog=True,
			vlim=[1e-3, 5],
			**kwargs
		)
		plot.run(
			'backscatter_hist',
			stats_clear_fine_filename,
			backscatter_hist_clear_fine_filename,
			vlog=True,
			vlim=[1e-3, 5],
			**kwargs
		)



import numpy as np
from alcf.algorithms import interp
from alcf import misc

def calc_filter_mask(filters, time, l):
	n = len(time)
	if len(filters) == 0:
		filter_mask = np.zeros(n, dtype=bool)
	else:
		filter_mask = np.ones(n, dtype=bool)
	for filter_ in filters:
		filter_mask_tmp = np.zeros(n, dtype=bool)
		m = filter_.shape[0]
		for i in range(m):
			t1, t2 = filter_[i]
			filter_mask_tmp |= (time >= t1) & (time < t2)
		filter_mask &= filter_mask_tmp
	if l > 0: filter_mask = np.tile(filter_mask, [l, 1]).T
	return filter_mask

def stats_map(d, state,
	tlim=None,
	blim=None,
	bres=None,
	bsd_lim=None,
	bsd_log=None,
	bsd_res=None,
	bsd_z=None,
	filter=None,
	filters_exclude=None,
	filters_include=None,
	zlim=None,
	zres=None,
	**kwargs
):
	if zlim is not None and zres is not None:
		state['zfull2'] = state.get('zfull2',
			np.arange(zlim[0] + 0.5*zres, zlim[1], zres)
		)
	else:
		state['zfull2'] = state.get('zfull2', d['zfull'])
	zhalf = misc.half(d['zfull'])
	zhalf2 = misc.half(state['zfull2'])
	state['backscatter_half'] = state.get('backscatter_half',
		np.arange(blim[0], blim[1] + bres, bres)
	)
	state['backscatter_full'] = state.get('backscatter_full',
		0.5*(state['backscatter_half'][1:] + state['backscatter_half'][:-1])
	)
	o = len(state['backscatter_full'])
	state['backscatter_sd_half'] = state.get('backscatter_sd_half',
		np.exp(np.arange(np.log(bsd_lim[0]), np.log(bsd_lim[1] + bsd_res),
			np.log(bsd_lim[0] + bsd_res) - np.log(bsd_lim[0])))
		if bsd_log is True \
		else np.arange(bsd_lim[0], bsd_lim[1] + bsd_res, bsd_res)
	)
	state['backscatter_sd_full'] = state.get('backscatter_sd_full',
		0.5*(state['backscatter_sd_half'][1:] + state['backscatter_sd_half'][:-1])
	)
	osd = len(state['backscatter_sd_full'])
	m2 = len(state['zfull2'])
	if len(d['cloud_mask'].shape) == 3:
		n, m, l = d['cloud_mask'].shape
		dims = (m, l)
		dims2 = (m2, l)
		hist_dims = (o, m, l)
		hist_dims2 = (o, m2, l)
		sd_hist_dims = (osd, l)
		filter_mask_dims = (n, l)
	else:
		n, m = d['cloud_mask'].shape
		l = 0
		dims = (m,)
		dims2 = (m2,)
		hist_dims = (o, m)
		hist_dims2 = (o, m2)
		sd_hist_dims = (osd,)
		filter_mask_dims = (n,)
	state['n'] = state.get('n',
		0 if l == 0 else np.zeros(l, dtype=np.int64)
	)
	state['time_total'] = state.get('time_total',
		0 if l == 0 else np.zeros(l, dtype=np.int64)
	)
	state['backscatter_avg'] = state.get(
		'backscatter_avg',
		np.zeros(dims2, dtype=np.float64)
	)
	state['backscatter_mol_avg'] = state.get(
		'backscatter_mol_avg',
		np.zeros(dims2, dtype=np.float64)
	)
	state['cl'] = state.get(
		'cl',
		np.zeros(dims2, dtype=np.float64)
	)
	state['cbh'] = state.get(
		'cbh',
		np.zeros(dims2, dtype=np.float64)
	)
	state['clt'] = state.get(
		'clt',
		np.zeros(l, dtype=np.float64) if l > 0 else 0
	)
	state['backscatter_hist'] = state.get(
		'backscatter_hist',
		np.zeros(hist_dims2, dtype=np.float64)
	)
	state['backscatter_sd_hist'] = state.get(
		'backscatter_sd_hist',
		np.zeros(sd_hist_dims, dtype=np.float64)
	)
	backscatter_hist_tmp = np.zeros(hist_dims, dtype=np.float64)
	cl_tmp = np.zeros(dims, dtype=np.float64)
	clt_tmp = np.zeros(l, dtype=np.float64) if l > 0 else 0
	backscatter_avg_tmp = np.zeros(dims, dtype=np.float64)
	backscatter_mol_avg_tmp = np.zeros(dims, dtype=np.float64)
	if tlim is not None:
		mask = (d['time'] >= tlim[0]) & (d['time'] < tlim[1])
	else:
		mask = np.ones(n, dtype=bool)

	if l > 0:
		mask &= np.all(~np.isnan(d['backscatter']), axis=(1, 2))
	else:
		mask &= np.all(~np.isnan(d['backscatter']), axis=(1,))

	filter_mask = np.ones(filter_mask_dims, dtype=bool)
	if 'cloudy' in filter:
		filter_mask &= np.any(d['cloud_mask'], axis=1)
	if 'clear' in filter:
		filter_mask &= ~np.any(d['cloud_mask'], axis=1)
	if 'day' in filter:
		filter_mask_0 = misc.sun_altitude(d['time'], d['lon'], d['lat']) >= 0
		if l > 0: filter_mask_0 = np.tile(filter_mask_0, [l, 1]).T
		filter_mask &= filter_mask_0
	if 'night' in filter:
		filter_mask_0 = misc.sun_altitude(d['time'], d['lon'], d['lat']) < 0
		if l > 0: filter_mask_0 = np.tile(filter_mask_0, [l, 1]).T
		filter_mask &= filter_mask_0

	if filters_exclude is not None:
		filter_mask &= ~calc_filter_mask(filters_exclude, d['time'], l)

	if filters_include is not None:
		filter_mask &= calc_filter_mask(filters_include, d['time'], l)

	if not np.any(mask):
		return
	for j in range(m):
		if l > 0:
			for k in range(l):
				backscatter_hist_tmp[:,j,k] += np.histogram(
					d['backscatter'][filter_mask[:,k] & mask,j,k],
					bins=state['backscatter_half'])[0]
		else:
			backscatter_hist_tmp[:,j] += np.histogram(
				d['backscatter'][filter_mask & mask,j],
				bins=state['backscatter_half'])[0]

	jsd = np.argmin(np.abs(d['zfull'] - bsd_z))
	state['backscatter_sd_z'] = d['zfull'][jsd]

	if 'backscatter_sd' in d:
		if l > 0:
			for k in range(l):
				state['backscatter_sd_hist'][:,k] += np.histogram(
					d['backscatter_sd'][filter_mask[:,k] & mask,jsd,k],
					bins=state['backscatter_sd_half'])[0]
		else:
			state['backscatter_sd_hist'] += np.histogram(
				d['backscatter_sd'][filter_mask & mask,jsd],
				bins=state['backscatter_sd_half'])[0]

	for i in range(o):
		if l > 0:
			for k in range(l):
				state['backscatter_hist'][i,:,k] += interp(
					zhalf,
					backscatter_hist_tmp[i,:,k],
					zhalf2
				)
		else:
			state['backscatter_hist'][i,:] += interp(
				zhalf,
				backscatter_hist_tmp[i,:],
				zhalf2
			)
	for i in range(n):
		if not mask[i]:
			continue
		if l > 0:
			for k in range(l):
				if not filter_mask[i,k]:
					continue
				cl_tmp[:,k] += d['cloud_mask'][i,:,k]
				if np.isfinite(d['cbh'][i,k]):
					j = np.argmin(np.abs(state['zfull2'] - d['cbh'][i,k]))
					state['cbh'][j,k] += 1
				backscatter_avg_tmp[:,k] += d['backscatter'][i,:,k]
				if 'backscatter_mol' in d:
					backscatter_mol_avg_tmp[:,k] += d['backscatter_mol'][i,:]
				state['n'][k] += 1
				state['time_total'][k] += \
					24*60*60*(d['time_bnds'][i,1] - d['time_bnds'][i,0])
				state['clt'][k] += np.any(d['cloud_mask'][i,:,k])
		else:
			if not filter_mask[i]:
				continue
			cl_tmp[:] += d['cloud_mask'][i,:]
			if np.isfinite(d['cbh'][i]):
				j = np.argmin(np.abs(state['zfull2'] - d['cbh'][i]))
				state['cbh'][j] += 1
			backscatter_avg_tmp[:] += d['backscatter'][i,:]
			if 'backscatter_mol' in d:
				backscatter_mol_avg_tmp[:] += d['backscatter_mol'][i,:]
			state['n'] += 1
			state['time_total'] += \
				24*60*60*(d['time_bnds'][i,1] - d['time_bnds'][i,0])
			state['clt'] += np.any(d['cloud_mask'][i,:])
	if l > 0:
		for k in range(l):
			state['cl'][:,k] += interp(
				zhalf,
				cl_tmp[:,k],
				zhalf2
			)
			state['backscatter_avg'][:,k] += interp(
				zhalf,
				backscatter_avg_tmp[:,k],
				zhalf2
			)
			state['backscatter_mol_avg'][:,k] += interp(
				zhalf,
				backscatter_mol_avg_tmp[:,k],
				zhalf2
			)
	else:
		state['cl'] += interp(
			zhalf,
			cl_tmp,
			zhalf2
		)
		state['backscatter_avg'] += interp(
			zhalf,
			backscatter_avg_tmp,
			zhalf2
		)
		state['backscatter_mol_avg'] += interp(
			zhalf,
			backscatter_mol_avg_tmp,
			zhalf2
		)

def stats_reduce(state, bsd_z=None, **kwargs):
	if len(state['cl'].shape) == 2:
		for k in range(len(state['n'])):
			if state['n'][k] > 0:
				state['backscatter_hist'][:,:,k] /= state['n'][k]
				state['cl'][:,k] /= state['n'][k]
				state['cbh'][:,k] /= state['clt'][k]
				state['clt'][k] /= state['n'][k]
				state['backscatter_avg'][:,k] /= state['n'][k]
				state['backscatter_mol_avg'][:,k] /= state['n'][k]
	else:
		if state['n'] != 0:
			state['backscatter_hist'] /= state['n']
			state['cl'] /= state['n']
			state['cbh'] /= state['clt']
			state['clt'] /= state['n']
			state['backscatter_avg'] /= state['n']
			state['backscatter_mol_avg'] /= state['n']
	state['backscatter_sd_hist'] /= state['n']
	do = {
		'cl': 100.*state['cl'],
		'cbh': 100.*state['cbh'],
		'clt': 100.*state['clt'],
		'zfull': state['zfull2'],
		'n': state['n'],
		'time_total': state['time_total'],
		'backscatter_avg': state['backscatter_avg'],
		'backscatter_mol_avg': state['backscatter_mol_avg'],
		'backscatter_full': state['backscatter_full'],
		'backscatter_hist': state['backscatter_hist'],
		'backscatter_sd_hist': state['backscatter_sd_hist'],
		'backscatter_sd_full': state['backscatter_sd_full'],
		'backscatter_sd_z': state['backscatter_sd_z'],
	}
	do['.'] = {
		'zfull': {
			'.dims': ['zfull'],
			'long_name': 'altitude of model full-levels',
			'standard_name': 'height_above_reference_ellipsoid',
			'units': 'm',
		},
		'cl': {
			'.dims': ['zfull', 'column'] \
				if len(state['cl'].shape) == 2 \
				else ['zfull'],
			'long_name': 'cloud area fraction',
			'standard_name': 'cloud_area_fraction_in_atmosphere_layer',
			'units': '%',
		},
		'cbh': {
			'.dims': ['zfull', 'column'] \
				if len(state['cl'].shape) == 2 \
				else ['zfull'],
			'long_name': 'cloud base height',
			'units': '%',
		},
		'clt': {
			'.dims': ['column'] \
				if isinstance(state['clt'], np.ndarray) \
				else [],
			'long_name': 'total cloud fraction',
			'standard_name': 'cloud_area_fraction',
			'units': '%',
		},
		'n': {
			'.dims': ['column'] \
				if len(state['cl'].shape) == 2 \
				else [],
			'long_name': 'number of profiles',
			'units': '1',
		},
		'time_total': {
			'.dims': ['column'] \
				if len(state['cl'].shape) == 2 \
				else [],
			'long_name': 'total time',
			'standard_name': 'time',
			'units': 's',
		},
		'backscatter_avg': {
			'.dims': ['zfull', 'column'] \
				if len(state['cl'].shape) == 2 \
				else ['zfull'],
			'long_name': 'total attenuated volume backscattering coefficient average',
			'units': 'm-1 sr-1',
		},
		'backscatter_mol_avg': {
			'.dims': ['zfull', 'column'] \
				if len(state['cl'].shape) == 2 \
				else ['zfull'],
			'long_name': 'total attenuated molecular volume backscattering coefficient average',
			'units': 'm-1 sr-1',
		},
		'backscatter_full': {
			'.dims': ['backscatter_full'],
			'long_name': 'total attenuated volume backscattering coefficient',
			'units': 'm-1 sr-1',
		},
		'backscatter_hist': {
			'.dims': ['backscatter_full', 'zfull'] \
				if len(state['backscatter_hist'].shape) == 2 \
				else ['backscatter_full', 'zfull', 'column'],
			'long_name': 'total attenuated volume backscattering coefficient histogram',
			'units': '%',
		},
		'backscatter_sd_hist': {
			'.dims': ['backscatter_sd_full'],
			'.dims': ['backscatter_sd_full'] \
				if len(state['backscatter_sd_hist'].shape) == 1 \
				else ['backscatter_sd_full', 'column'],
			'long_name': 'total attenuated volume backscattering coefficient standard deviation histogram',
			'units': '%',
		},
		'backscatter_sd_full': {
			'.dims': ['backscatter_sd_full'],
			'long_name': 'total attenuated volume backscattering coefficient standard deviation',
			'units': 'm-1 sr-1',
		},
		'backscatter_sd_z': {
			'.dims': [],
			'long_name': 'total attenuated volume backscattering coefficient standard deviation height above reference ellipsoid',
			'units': 'm',
		},
	}
	return do

def stream(dd, state, **options):
	state['state'] = state.get('state', {})
	for d in dd:
		if d is None:
			return [stats_reduce(state['state'], **options)]
		stats_map(d, state['state'], **options)
	return []



import os
import copy
import tempfile
from string import Template
import subprocess
from alcf.lidars import LIDARS

CONFIG_TEMPLATE = """
&config_nml
	config%NPOINTS_IT=500,
	config%NCOLUMNS=${ncolumns},
	config%NLEVELS=${nlevels},
	config%USE_VGRID=.true.,
	config%NLR=40,
	config%CSAT_VGRID=.true.,
	config%RADAR_FREQ=94.0,
	config%SURFACE_RADAR=0,
	config%use_mie_tables=0,
	config%use_gas_abs=1,
	config%do_ray=0,
	config%melt_lay=0,
	config%k2=-1,
	config%use_reff=.true.,
	config%use_precipitation_fluxes=.false.,
	config%Nprmts_max_hydro=12,
	config%Naero=1,
	config%Nprmts_max_aero=1,
	config%lidar_ice_type=0,
	config%lidar_wavelength=${wavelength},
	config%lidar_max_range=${max_range},
	config%surface_lidar=${surface_lidar},
	config%OVERLAP=${overlap},
	config%ISCCP_TOPHEIGHT=1,
	config%ISCCP_TOPHEIGHT_DIRECTION=2,
	config%Platform=1,
	config%Satellite=15,
	config%Instrument=0,
	config%Nchannels=8,
	config%Channels=1,3,5,6,8,10,11,13,
	config%Surfem=0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
	config%ZenAng=50.0,
	config%CO2=5.241e-04,
	config%CH4=9.139e-07,
	config%N2O=4.665e-07,
	config%CO=2.098e-07,
	config%output%Lradar_sim=.false.,
	config%output%Llidar_sim=.true.,
	config%output%Lisccp_sim=.false.,
	config%output%Lmisr_sim=.false.,
	config%output%Lmodis_sim=.false.,
	config%output%Lrttov_sim=.false.,
	config%output%Ltoffset=.true.,
	config%output%Lcfaddbze94=.false.,
	config%output%Ldbze94=.false.,
	config%output%Latb532=.false.,
	config%output%LcfadLidarsr532=.false.,
	config%output%Lclcalipso=.false.,
	config%output%Lclhcalipso=.false.,
	config%output%Lcllcalipso=.false.,
	config%output%Lclmcalipso=.false.,
	config%output%Lcltcalipso=.false.,
	config%output%LparasolRefl=.false.,
	config%output%Lclcalipsoliq=.false.,
	config%output%Lclcalipsoice=.false.,
	config%output%Lclcalipsoun=.false.,
	config%output%Lclcalipsotmp=.false.,
	config%output%Lclcalipsotmpliq=.false.,
	config%output%Lclcalipsotmpice=.false.,
	config%output%Lclcalipsotmpun=.false.,
	config%output%Lclhcalipsoliq=.false.,
	config%output%Lcllcalipsoliq=.false.,
	config%output%Lclmcalipsoliq=.false.,
	config%output%Lcltcalipsoliq=.false.,
	config%output%Lclhcalipsoice=.false.,
	config%output%Lcllcalipsoice=.false.,
	config%output%Lclmcalipsoice=.false.,
	config%output%Lcltcalipsoice=.false.,
	config%output%Lclhcalipsoun=.false.,
	config%output%Lcllcalipsoun=.false.,
	config%output%Lclmcalipsoun=.false.,
	config%output%Lcltcalipsoun=.false.,
	config%output%Lalbisccp=.false.,
	config%output%Lboxptopisccp=.false.,
	config%output%Lboxtauisccp=.false.,
	config%output%Lpctisccp=.false.,
	config%output%Lclisccp=.false.,
	config%output%Ltauisccp=.false.,
	config%output%Lcltisccp=.false.,
	config%output%Lmeantbisccp=.false.,
	config%output%Lmeantbclrisccp=.false.,
	config%output%LclMISR=.false.,
	config%output%Lclcalipso2=.false.,
	config%output%Lcltlidarradar=.false.,
	config%output%Lfracout=.false.,
	config%output%LlidarBetaMol532=.false.,
	config%output%Lcltmodis=.true.,
	config%output%Lclwmodis=.true.,
	config%output%Lclimodis=.true.,
	config%output%Lclhmodis=.true.,
	config%output%Lclmmodis=.true.,
	config%output%Lcllmodis=.true.,
	config%output%Ltautmodis=.true.,
	config%output%Ltauwmodis=.true.,
	config%output%Ltauimodis=.true.,
	config%output%Ltautlogmodis=.true.,
	config%output%Ltauwlogmodis=.true.,
	config%output%Ltauilogmodis=.true.,
	config%output%Lreffclwmodis=.true.,
	config%output%Lreffclimodis=.true.,
	config%output%Lpctmodis=.true.,
	config%output%Llwpmodis=.true.,
	config%output%Liwpmodis=.true.,
	config%output%Lclmodis=.true.,
	config%output%Lcrimodis=.true.,
	config%output%Lcrlmodis=.true.,
	config%output%Ltbrttov=.false.,
/
"""

OVERLAP = {
	'maximum': 1,
	'random': 2,
	'maximum-random': 3,
}

def cosp_alcf(config, input_, output):
	_, config_filename = tempfile.mkstemp('.nml', prefix='alcf_config_', text=False)
	try:
		with open(config_filename, 'w') as f:
			f.write(config)
		program = os.path.join(os.path.dirname(__file__), '../cosp_alcf')
		subprocess.call([program, config_filename, input_, output])
	finally:
		os.unlink(config_filename)

def run(type_, input_, output,
	ncolumns=10,
	overlap='maximum-random',
	**kwargs
):
	'''
alcf-simulate -- Simulate lidar measurements from model data using COSP.
=============

Synopsis
--------

    alcf simulate <type> [<options>] [--] <input> <output>

Description
-----------

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `type`: Type of lidar to simulate.
- `input`: Input filename or directory (the output of "alcf model").
- `output`: Output filename or directory.
- `options`: See Options below.

Types
-----

- `caliop`: CALIPSO/CALIOP.
- `chm15k`: Lufft CHM 15k.
- `cl31`: Vaisala CL31.
- `cl51`: Vaisala CL51.
- `cl61`: Vaisala CL61.
- `mpl`: Sigma Space MiniMPL.

Options
-------

- `ncolumns: <ncolumns>`: Number of SCOPS subcolumns to generate. Default: `10`.
- `overlap: <overlap>`: Cloud overlap assumption in the SCOPS subcolumn generator. `maximum` for maximum overlap, `random` for random overlap, or `maximum-random` for maximum-random overlap. Default: `maximum-random`.

Examples
--------

Simulate a Vaisala CL51 instrument from model data in `alcf_merra2_model` previously extracted using `alcf model` and store the output in the direcctory `alcf_merra2_simulate`.

    alcf simulate cl51 alcf_merra2_model alcf_merra2_simulate
	'''
	lidar = LIDARS.get(type_)
	if lidar is None:
		raise ValueError('Invalid type: %s' % type_)

	overlap_flag = OVERLAP.get(overlap)
	if overlap_flag is None:
		raise ValueError('Invalid overlap: %s' % overlap)

	nlevels = 60

	template = Template(CONFIG_TEMPLATE)
	config = template.substitute(
		ncolumns=ncolumns,
		nlevels=nlevels,
		overlap=overlap_flag,
		wavelength=lidar.WAVELENGTH,
		max_range=lidar.MAX_RANGE,
		surface_lidar=(1 if lidar.SURFACE_LIDAR else 0),
	)

	if os.path.isfile(input_):
		print('<- %s' % input_)
		cosp_alcf(config, input_, output)
	else:
		files = sorted(os.listdir(input_))
		for file_ in files:
			input_filename = os.path.join(input_, file_)
			output_filename = os.path.join(output, file_)
			if not os.path.isfile(input_filename):
				continue
			print('<- %s' % input_filename)
			cosp_alcf(config, input_filename, output_filename)
			print('-> %s' % output_filename)



import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc

VARIABLES = [
	'hybridt32',
	'latitude',
	'longitude',
	'model_press',
	'model_qcf',
	'model_qcl',
	'theta_lev_temp',
	'time0',
]

STEP = 2/24

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	dd_index = ds.readdir(dirname, variables=['time0', 'latitude', 'longitude'],
		jd=True, recursive=recursive)
	dd = []
	for d_index in dd_index:
		time = d_index['time0']
		lon = d_index['longitude']
		lon = np.where(lon < 0., 360. + lon, lon)
		lat = d_index['latitude']
		filename = d_index['filename']
		ii = np.where((time >= t1 - step*0.5) & (time <= t2 + step*0.5))[0]
		for i in ii:
			lon0, lat0 = track(time[i])
			if np.isnan(lon0) or np.isnan(lat0):
				continue
			l = np.argmin((lon - lon0)**2 + (lat - lat0)**2)
			j, k = np.unravel_index(l, lon.shape)
			# print('<- %s' % filename)
			d = ds.read(filename, variables=VARIABLES, sel={'time0': i, 'rlat': j, 'rlon': k})
			if not set(VARIABLES).issubset(d.keys()):
				continue
			clw = d['model_qcl']
			cli = d['model_qcf']
			cl = 100.*np.ones(len(clw), dtype=np.float64)
			ps = 2*d['model_press'][0] - d['model_press'][1]
			orog = max(0., 2*d['hybridt32'][0] - d['hybridt32'][1])
			pfull = d['model_press']
			zfull = d['hybridt32']
			ta = d['theta_lev_temp']
			newshape4 = (1, len(clw))
			newshape3 = (1,)
			d_new = {
				'clw': clw.reshape(newshape4),
				'cli': cli.reshape(newshape4),
				'ta': ta.reshape(newshape4),
				'cl': cl.reshape(newshape4),
				'pfull': pfull.reshape(newshape4),
				'zfull': zfull.reshape(newshape4),
				'ps': [ps],
				'orog': [orog],
				'lon': np.array([lon[j,k]]),
				'lat': np.array([lat[j,k]]),
				'time': np.array([time[i]]),
				'.': META,
			}
			dd.append(d_new)
	d = ds.op.merge(dd, 'time')
	return d



import sys
import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc
import aquarius_time as aq

VARS_PLEV = [
	'time',
	'clwc',
	'ciwc',
	'cc',
	'latitude',
	'longitude',
	'level',
	't',
	'z',
]

VARS_SURF = [
	'time',
	'sp',
	'z',
	'latitude',
	'longitude',
]

TRANS_PLEV = {
	'latitude': 'lat',
	'longitude': 'lon',
	'z': 'zfull',
	'cc': 'cl',
	'level': 'pfull',
	'clwc': 'clw',
	'ciwc': 'cli',
	't': 'ta',
}

TRANS_SURF = {
	'latitude': 'lat',
	'longitude': 'lon',
	'z': 'orog',
	'sp': 'ps',
}

STEP = 1/24

def read0(type_, dirname, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	dd_idx = ds.readdir(dirname,
		variables=['time', 'latitude', 'longitude'],
		jd=True,
		full=True,
		warnings=warnings,
		recursive=recursive,
	)

	vars = {
		'surf': VARS_SURF,
		'plev': VARS_PLEV,
	}[type_]

	trans = {
		'surf': TRANS_SURF,
		'plev': TRANS_PLEV,
	}[type_]

	dd = []
	for d_idx in dd_idx:
		time = d_idx['time']
		lat = d_idx['latitude']
		lon = d_idx['longitude']
		lon = lon % 360
		filename = d_idx['filename']

		ii = np.nonzero(
			(time >= t1 - step*0.5) &
			(time < t2 + step*0.5)
		)[0]
		for i in ii:
			t = time[i]
			lon0, lat0 = track(time[i])
			if np.isnan(lon0) or np.isnan(lat0):
				continue
			j = np.argmin(np.abs(lat - lat0))
			k = np.argmin(np.abs(lon - lon0))
			d = ds.read(filename, vars,
				sel={'time': [i], 'latitude': j, 'longitude': k},
				jd=True,
			)
			for a, b in trans.items():
				if a in d.keys():
					ds.rename(d, a, b)
			d['lat'] = np.array([d['lat']])
			d['lon'] = np.array([d['lon']])
			d['lon'] = d['lon'] % 360
			d['.']['lat']['.dims'] = ['time']
			d['.']['lon']['.dims'] = ['time']
			if type_ == 'plev':
				d['pfull'] = d['pfull'].reshape([1, len(d['pfull'])])
				d['.']['pfull']['.dims'] = ['time', 'level']
				d['cl'] = d['cl'][:,::-1]
				d['clw'] = d['clw'][:,::-1]
				d['cli'] = d['cli'][:,::-1]
				d['ta'] = d['ta'][:,::-1]
				d['zfull'] = d['zfull'][:,::-1]
				d['pfull'] = d['pfull'][:,::-1]
			dd.append(d)
	d = ds.op.merge(dd, 'time')
	if 'pfull' in d:
		d['pfull'] = 1e2*d['pfull']
	if 'zfull' in d:
		d['zfull'] /= 9.80665
	if 'orog' in d:
		d['orog'] /= 9.80665
	if 'cl' in d:
		d['cl'] = np.minimum(1, np.maximum(0, d['cl']))
		d['cl'] *= 100.
	return d

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	d_surf = read0('surf', os.path.join(dirname, 'surf'), track, t1, t2,
		warnings=warnings, step=step, recursive=recursive)
	d_plev = read0('plev', os.path.join(dirname, 'plev'), track, t1, t2,
		warnings=warnings, step=step, recursive=recursive)
	d = {**d_surf, **d_plev}
	d['.'] = META
	return d



from alcf import misc

def calibration(d, calibration_coeff=1.0, **options):
	if 'backscatter' in d:
		d['backscatter'] *= calibration_coeff
	if 'backscatter_mol' in d:
		d['backscatter_mol'] *= calibration_coeff
	if 'backscatter_sd' in d:
		d['backscatter_sd'] *= calibration_coeff

def stream(dd, state, **options):
	return misc.stream(dd, state, calibration, **options)



import copy
import numpy as np
import ds_format as ds
import aquarius_time as aq
from alcf import misc
from alcf.algorithms import interp

VARIABLES = [
	'zfull',
	'backscatter_sd',
	'backscatter_mol',
]

def couple(d, d_idx):
	dims = d['backscatter'].shape
	n = dims[0]
	l = dims[2] if len(dims) == 3 else 0
	couple_bsd = False
	couple_bmol = False
	if 'backscatter_sd' not in d:
		couple_bsd = True
		d['backscatter_sd'] = np.full(dims, np.nan, np.float64)
		d['.']['backscatter_sd'] = {
			'.dims': ['time', 'level', 'column'] \
				if len(dims) == 3 \
				else ['time', 'level'],
			'long_name': 'total_attenuated_backscatter_coefficient_standard_deviation',
			'units': 'm-1 sr-1',
		}
	if 'backscatter_mol' not in d:
		couple_bmol = True
		d['backscatter_mol'] = np.full(dims, np.nan, np.float64)
		d['.']['backscatter_mol'] = {
			'.dims': ['time', 'level'],
			'long_name': 'total_attenuated_molecular_backscatter_coefficient',
			'units': 'm-1 sr-1',
		}
	for i in range(n):
		t = d['time'][i]
		j = np.argmin(np.abs(d_idx['time'] - t))
		n1 = d_idx['n'][j]
		filename = d_idx['filename'][n1]
		i1 = d_idx['i'][j]
		d1 = ds.read(filename, VARIABLES, {'time': i1})
		zhalf1 = misc.half(d1['zfull'])
		zhalf = misc.half(d['zfull'][i,:]) \
			if d['zfull'].ndim == 2 \
			else misc.half(d['zfull'])
		if couple_bsd and 'backscatter_sd' in d1:
			b_sd = interp(zhalf1, d1['backscatter_sd'], zhalf)
			if len(dims) == 3:
				for k in range(l):
					d['backscatter_sd'][i,:,k] = b_sd
			else:
				d['backscatter_sd'][i,:] = b_sd
		if couple_bmol and 'backscatter_mol' in d1:
			b_mol = interp(zhalf1, d1['backscatter_mol'], zhalf)
			d['backscatter_mol'][i,:] = b_mol

def stream(dd, state, dirname):
	if 'd_idx' not in state:
		d_idx = ds.readdir(dirname, ['time'], merge='time')
		state['d_idx'] = d_idx
	return misc.stream(dd, state, couple, d_idx=state['d_idx'])




import sys
import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc
import aquarius_time as aq

VARS = [
	'gh',
	't',
	'ciwc',
	'clw',
	'tcc',
	'sp',
]

VARS_AUX = [
	'level',
	'time',
	'latitude',
	'longitude',
]

TRANS = {
	'gh': 'zfull',
	'latitude': 'lat',
	'longitude': 'lon',
	'level': 'pfull',
	't': 'ta',
	'ciwc': 'cli',
	'clw': 'clw',
	'tcc': 'cl',
	'time': 'time',
	'sp': 'ps',
}

STEP = 6/24

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	d_ll = ds.read(os.path.join(dirname, 'LL125.nc'), [
		'latitude',
		'longitude',
		'z'
	])
	lat_ll = d_ll['latitude']
	lon_ll = d_ll['longitude']
	orog_ll = d_ll['z'][0,:,:]/9.80665

	dd_idx = ds.readdir(dirname,
		variables=['time', 'latitude', 'longitude'],
		jd=True,
		full=True,
		warnings=warnings,
		recursive=recursive,
	)
	d_out = {}
	for var in VARS:
		dd = []
		var2 = TRANS[var]
		for d_idx in dd_idx:
			if var not in d_idx['.']:
				continue
			time = d_idx['time']
			lat = d_idx['latitude']
			lon = d_idx['longitude']
			filename = d_idx['filename']
			ii = np.nonzero(
				(time >= t1 - step*0.5) &
				(time < t2 + step*0.5)
			)[0]
			for i in ii:
				t = time[i]
				lon0, lat0 = track(time[i])
				if np.isnan(lon0) or np.isnan(lat0):
					continue
				j = np.argmin(np.abs(lat - lat0))
				k = np.argmin(np.abs(lon - lon0))
				j_ll = np.argmin(np.abs(lat_ll - lat0))
				k_ll = np.argmin(np.abs(lon_ll - lon0))
				d = ds.read(filename, VARS_AUX + [var],
					sel={
						'time': [i],
						'latitude': j,
						'longitude': k,
					},
					jd=True,
				)
				for a, b in TRANS.items():
					if a in d.keys():
						ds.rename(d, a, b)
				d['lat'] = np.array([d['lat']])
				d['lon'] = np.array([d['lon']])
				d['orog'] = np.array([orog_ll[j_ll,k_ll]])
				d['.']['lat']['.dims'] = ['time']
				d['.']['lon']['.dims'] = ['time']
				d['.']['orog'] = {'.dims': ['time']}
				if 'pfull' in ds.get_vars(d):
					d['pfull'] = d['pfull'].reshape([1, len(d['pfull'])])
					d['.']['pfull']['.dims'] = ['time', 'pfull']
					d['pfull'] = d['pfull'][:,::-1]
					d[var2] = d[var2][:,::-1]
					ds.select(d, {'pfull': np.arange(27)})
				dd.append(d)
		d = ds.op.merge(dd, 'time')
		for var_aux in VARS_AUX:
			if TRANS[var_aux] in ds.get_vars(d_out) \
				and TRANS[var_aux] in ds.get_vars(d) \
				and not np.all(d_out[TRANS[var_aux]] == d[TRANS[var_aux]]):
				raise ValueError('%s: Field differs between input files' % TRANS[var_aux])
		d_out.update(d)
	d_out['pfull'] = d_out['pfull']*1e2
	return d_out



import sys
import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc
import aquarius_time as aq

VARS = [
	'TALLTS',
	'latitude_t',
	'longitude_t',
	'DALLTH_eta_theta',
	'STASH_m01s00i265',
	'STASH_m01s00i408',
	'STASH_m01s00i409',
	'STASH_m01s00i254',
	'STASH_m01s00i012',
	'STASH_m01s16i004',
]

TRANS = {
	'TALLTS': 'time',
	'latitude_t': 'lat',
	'longitude_t': 'lon',
	'DALLTH_eta_theta': 'eta',
	'STASH_m01s00i265': 'cl',
	'STASH_m01s00i408': 'pfull',
	'STASH_m01s00i409': 'ps',
	'STASH_m01s00i254': 'clw',
	'STASH_m01s00i012': 'cli',
	'STASH_m01s16i004': 'ta',
}

STEP = 1/24

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	d_orog = ds.read(os.path.join(dirname, 'qrparm.orog.nc'), [
		'latitude',
		'longitude',
		'surface_altitude',
	])

	dd_idx = ds.readdir(dirname,
		variables=['TALLTS', 'latitude_t', 'longitude_t', 'DALLTH_zsea_theta'],
		jd=True,
		full=True,
		warnings=warnings,
		recursive=recursive,
	)

	dd = []
	for d_idx in dd_idx:
		if 'TALLTS' not in d_idx:
			continue

		time = d_idx['TALLTS']
		lat = d_idx['latitude_t']
		lon = d_idx['longitude_t']
		filename = d_idx['filename']

		ii = np.nonzero(
			(time >= t1 - step*0.5) &
			(time < t2 + step*0.5)
		)[0]
		for i in ii:
			t = time[i]
			lon0, lat0 = track(time[i])
			if np.isnan(lon0) or np.isnan(lat0):
				continue
			j = np.argmin(np.abs(lat - lat0))
			k = np.argmin(np.abs(lon - lon0))
			d = ds.read(filename, VARS,
				sel={'TALLTS': [i], 'latitude_t': j, 'longitude_t': k},
				jd=True,
			)
			for a, b in TRANS.items():
				if a in d.keys():
					ds.rename(d, a, b)
			d['lat'] = np.array([d['lat']])
			d['lon'] = np.array([d['lon']])
			d['.']['lat']['.dims'] = ['time']
			d['.']['lon']['.dims'] = ['time']
			orog = d_orog['surface_altitude'][j,k]
			d['zfull'] = d['eta']*85000. + orog*(1. - d['eta']/d['eta'][51])**2
			d['zfull'] = d['zfull'].reshape([1, len(d['zfull'])])
			d['.']['zfull'] = {'.dims': ['time', 'level']}
			d['orog'] = np.array([orog], np.float64)
			d['.']['orog'] = {'.dims': ['time']}
			del d['eta']
			dd.append(d)
	d = ds.op.merge(dd, 'time')
	d['cl'] *= 100.
	return d



import numpy as np
import ds_format as ds
from alcf import misc
from alcf.lidars import META

WAVELENGTH = 1064 # nm
CALIBRATION_COEFF = 1.0
SURFACE_LIDAR = None
SC_LR = None

def read(filename, vars, altitude=None, lon=None, lat=None, time=None, **kwargs):
	sel = None
	if time is not None:
		d = ds.read(filename, 'time_bnds', jd=True)
		mask = misc.time_mask(d['time_bnds'], time[0], time[1])
		if np.sum(mask) == 0: return None
		sel = {'time': mask}

	d = ds.read(filename, vars, sel=sel)
	n = d['backscatter'].shape[0]
	d['altitude'] = d['altitude'] if altitude is None and 'altitude' in d else \
		np.full(n, altitude, np.float64)
	d['lon'] = d['lon'] if lon is None and 'lon' in d else \
		np.full(n, lon, np.float64)
	d['lat'] = d['lat'] if lat is None and 'lat' in d else \
		np.full(n, lat, np.float64)
	d['.']['altitude'] = META['altitude']
	d['.']['lon'] = META['lon']
	d['.']['lat'] = META['lat']
	d['.'] = {
		x: d['.'][x]
		for x in vars
		if x in d['.']
	}
	return d




import os
from alcf.cmds import lidar, stats, plot

def run(*args, **options):
	input_ = args[:-1]
	output = args[-1]
	stats_dirs = [
		os.path.join(dir, 'stats')
		for dir in input_
	]
	stats_filenames = [
		os.path.join(stats_dir, 'all.nc')
		for stats_dir in stats_dirs
	]
	options['labels'] = options.get('labels',
		[os.path.dirname(dir) for dir in input_]
	)
	options['title'] = options.get('title',
		os.path.basename(output)
	)
	plot_dir = os.path.join(output, 'plot')
	cloud_occurrence_filename = os.path.join(plot_dir, 'cloud_occurrence.png')

	print('-> %s' % output)
	try: os.mkdir(output)
	except OSError: pass
	print('-> %s' % plot_dir)
	try: os.mkdir(plot_dir)
	except OSError: pass
	print('! alcf plot cloud_occurrence')
	print(options)
	plot.run('cloud_occurrence', *stats_filenames, cloud_occurrence_filename,
		**options
	)



import sys
import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc
import aquarius_time as aq

VARS = [
	'cli',
	'clw',
	'pfull',
	'phalf',
	'ta',
	'zg',
	'zghalf',
]

STEP = 3/24

def index(dirname, warnings=[], recursive=False, njobs=1):
	import intake
	try: path, model, run, timestep, zoom = dirname
	except Exception as e:
		raise ValueError('Invalid input format: "%s"' % dirname) from e
	cat = intake.open_catalog(path)
	obj = getattr(getattr(cat, model), run)
	ids = obj(time=timestep, zoom=zoom).to_dask()
	time = aq.from_iso('1970-01-01') + \
		(np.array(ids.time).astype('datetime64[s]').astype('int'))/(24*60*60)
	return {
		'ids': ids,
		'time': time,
	}

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	import healpy
	ids = index['ids']
	time = index['time']
	nest = ids.crs.healpix_order == 'nest'
	ii = np.nonzero(
		(time >= t1 - step*0.5) &
		(time < t2 + step*0.5)
	)[0]
	dd = []
	for i in ii:
		t = time[i]
		lon, lat = track(t)
		if np.isnan(lon) or np.isnan(lat):
			continue
		j = healpy.ang2pix(ids.crs.healpix_nside, lon, lat,
			lonlat=True,
			nest=nest
		)
		d = {}
		d['time'] = t
		d['lon'], d['lat'] = healpy.pix2ang(ids.crs.healpix_nside, j,
			lonlat=True,
			nest=nest
		)
		for var in VARS:
			sel = {'cell': j}
			if 'time' in ids[var].coords:
				sel['time'] = i
			x = np.array(ids[var].isel(**sel))
			if var == 'phalf':
				d['ps'] = x[-1]
				dims = []
			elif var == 'zghalf':
				d['orog'] = x[-1]
				dims = []
			elif var == 'zg':
				d['zfull'] = x[::-1]
				dims = ['level']
			else:
				d[var] = x[::-1]
				dims = ['level']
			ds.dims(d, var, dims)
		dd += [d]
	d = ds.merge(dd, 'time')
	d['cl'] = np.full(d['cli'].shape, 100., np.float64)
	return d



import ds_format as ds
import os
import numpy as np
from alcf.models import META
from alcf import misc

VARS = [
	'lon',
	'lat',
	'time',
	'H',
	'T',
	'PL',
	'PS',
	'PHIS',
	'QL',
	'QI',
	'CLOUD',
]

STEP=3/24

def read(dirname, index, track, t1, t2,
	warnings=[], step=STEP, recursive=False):

	dd_index = ds.readdir(dirname, variables=['time', 'lat', 'lon'], jd=True,
		recursive=recursive)
	dd = []
	for d_index in dd_index:
		time = d_index['time']
		lat = d_index['lat']
		lon = d_index['lon']
		lon = np.where(lon < 0., 360. + lon, lon)
		filename = d_index['filename']
		ii = np.nonzero(
			(time >= t1 - step*0.5) &
			(time < t2 + step*0.5)
		)[0]
		for i in ii:
			t = time[i]
			lon0, lat0 = track(time[i])
			if np.isnan(lon0) or np.isnan(lat0):
				continue
			j = np.argmin(np.abs(lat - lat0))
			k = np.argmin(np.abs(lon - lon0))
			d = ds.read(filename, variables=VARS,
				sel={'time': i, 'lat': j, 'lon': k}
			)
			clw = d['QL'][::-1]
			cli = d['QI'][::-1]
			cl = d['CLOUD'][::-1]*100.
			ps = d['PS']
			orog = d['PHIS']/9.80665
			pfull = d['PL'][::-1]
			zfull = d['H'][::-1]
			ta = d['T'][::-1]
			nlev = len(clw)
			newshape4 = (1, nlev)
			newshape3 = (1,)
			d_new = {
				'clw': clw.reshape(newshape4),
				'cli': cli.reshape(newshape4),
				'ta': ta.reshape(newshape4),
				'cl': cl.reshape(newshape4),
				'pfull': pfull.reshape(newshape4),
				'zfull': zfull.reshape(newshape4),
				'ps': ps.reshape(newshape3),
				'orog': orog.reshape(newshape3),
				'lat': np.array([lat[j]]),
				'lon': np.array([lon[k]]),
				'time': np.array([t]),
				'.': META,
			}
			dd.append(d_new)
	d = ds.op.merge(dd, 'time')
	return d



import numpy as np
import ds_format as ds
from alcf import misc
from alcf.lidars import META

WAVELENGTH = 1064 # nm
CALIBRATION_COEFF = 0.34
SURFACE_LIDAR = True
SC_LR = 18.2 # sr. Stratocumulus lidar ratio (O'Connor et al., 2004).
MAX_RANGE = 15400 # m

VARS = {
	'backscatter': ['beta_raw'],
	'zfull': ['range'],
}

DEFAULT_VARS = [
	'time',
	'altitude'
]

def read(
	filename,
	vars,
	altitude=None,
	lon=None,
	lat=None,
	tlim=None,
	keep_vars=[],
	**kwargs
):
	sel = None
	tres = None
	if tlim is not None:
		d = ds.read(filename, 'time', jd=True)
		tres = d['time'][1] - d['time'][0]
		d['time_bnds'] = misc.time_bnds(d['time'], tres)
		mask = misc.time_mask(d['time_bnds'], tlim[0], tlim[1])
		if np.sum(mask) == 0: return None
		sel = {'time': mask}

	dep_vars = misc.dep_vars(VARS, vars)
	req_vars = dep_vars + DEFAULT_VARS + keep_vars
	d = ds.read(filename, req_vars, jd=True, sel=sel, full=True)
	dx = {}
	misc.populate_meta(dx, META, set(vars) & set(VARS))
	n = ds.dim(d, 'time')
	m = ds.dim(d, 'range')
	if altitude is None:
		altitude = d['altitude']
	if 'time' in vars:
		dx['time'] = d['time']
	if 'time_bnds' in vars:
		if tres is None: tres = d['time'][1] - d['time'][0]
		args = [] if tlim is None else [tlim[0], tlim[1]]
		dx['time_bnds'] = misc.time_bnds(d['time'], tres, *args)
	if 'backscatter' in vars:
		dx['backscatter'] = d['beta_raw']*1e-11*CALIBRATION_COEFF
	if 'zfull' in vars:
		zfull1 = d['range'] + altitude
		dx['zfull'] = np.tile(zfull1, (n, 1))
	if 'altitude' in vars:
		dx['altitude'] = np.full(n, altitude, np.float64)
	if 'lon' in vars:
		dx['lon'] = np.full(n, lon, np.float64)
	if 'lat' in vars:
		dx['lat'] = np.full(n, lat, np.float64)
	for var in keep_vars:
		misc.keep_var(var, d, dx)
	return dx



import sys
import os
import traceback
from concurrent.futures import ProcessPoolExecutor, as_completed
import numpy as np
import aquarius_time as aq
import ds_format as ds
from alcf.models import MODELS, META
from alcf import misc

def override_year_in_time(time, year):
	try: len(time)
	except:	return override_year_in_time(np.array([time]), year)[0]
	date = aq.to_date(time)
	y = date[1]
	n = len(y)
	if np.all(y == year):
		return time
	ones = np.ones(n, int)
	zeros = np.zeros(n, int)
	start_old = aq.from_date([ones, y, ones, ones, zeros, zeros, zeros, zeros])
	start_new_1 = aq.from_date([1, year, 1, 1, 0, 0, 0])
	start_new = np.full(n, start_new_1)
	dt = time - start_old
	time_new = start_new + dt
	# Do this again in case the day overflows because of the old year is a leap
	# year while the new is not, and the time as near the end of the year.
	return override_year_in_time(time_new, year)

def worker(type_, input_, index, output, track, start, debug, r,
	override_year=None):
	try:
		if override_year is not None:
			t1 = override_year_in_time(start, override_year)
		else:
			t1 = start
		t2 = t1 + 1

		def track_f(t):
			dt = t - t1
			return misc.track_at(track, start + dt)

		model = MODELS[type_]
		output_filename = os.path.join(output, '%s.nc' % \
			aq.to_iso(start).replace(':', ''))
		warnings = []
		d = model.read(input_, index, track_f, t1, t2,
			warnings=warnings,
			recursive=r,
		)
		for w in warnings:
			if len(w) == 2:
				print('Warning: %s' % w[0], file=sys.stderr)
				if debug: print(w[1], file=sys.stderr)
				else: print('Use --debug to print debugging information.', file=sys.stderr)
			else:
				print('Warning: %s' % w, file=sys.stderr)
		if d is not None:
			if 'time_bnds' not in d and 'time' in d:
				d['time_bnds'] = misc.time_bnds(d['time'], model.STEP, t1, t2)
			for var in ['time', 'time_bnds']:
				if 'time' in d:
					d[var] = start + (d[var] - t1)
			d['.'] = META
			ds.write(output_filename, d)
			print('-> %s' % output_filename)
	except Exception as e:
		print('Warning: %s' % str(e), file=sys.stderr)
		if debug: print(traceback.format_exc(), file=sys.stderr)
		else: print('Use --debug to print debugging information.', file=sys.stderr)

def run(type_, input_, output,
	point=None,
	time=None,
	track=None,
	track_gap=21600,
	override_year=None,
	track_lon_180=False,
	debug=False,
	r=False,
	njobs=None,
	**kwargs
):
	'''
alcf-model -- Extract model data at a point or along a track.
==========

Synopsis
--------

    alcf model <type> point: { <lon> <lat> } time: { <start> <end> } [options] [--] <input> <output>

    alcf model <type> track: <track> [options] [--] <input> <output>

Description
-----------

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `type`: Input data type (see Types below).
- `input`: Input directory.
- `output`: Output directory.
- `lon`: Point longitude (degrees East).
- `lat`: Point latitutde (degrees North).
- `start`: Start time (see Time format below).
- `end`: End time (see Time format below).
- `track: <file>`, `track: { <file>... }`: One or more track NetCDF files (see Files below). If multiple files are supplied and `time_bnds` is not present in the files, they are assumed to be multiple segments of a discontinous track unless the last and first time of adjacent tracks are the same.
- `options`: See Options below.

Options
-------

- `njobs: <n>`: Number of parallel jobs. Default: number of CPU cores.
- `-r`: Process the input directory recursively.
- `--track_lon_180`: Expect track longitude between -180 and 180 degrees. This option is no longer needed as the conversion is automatically. [deprecated]
- `override_year: <year>`: Override year in the track. Use if comparing observations with a model statistically and the model output does not have a corresponding year available. The observation time is converted to the same time relative to the start of the year in the specified year. Note that if the original year is a leap year and the override year is not, as a consequence of the above 31 December is mapped to 1 January. The output retains the original year as in the track, even though the model data come from the override year. Default: `none`.
- `track_gap: <interval>`: If the interval is not 0, a track file is supplied, the `time_bnds` variable is not defined in the file and any two adjacent points are separated by more than the specified time interval (seconds), then a gap is assumed to be present between the two data points, instead of interpolating location between the two points. Default: `21600` (6 hours).

Types
-----

- `amps`: Antarctic Mesoscale Prediction System (AMPS).
- `era5`: ERA5.
- `icon`: ICON.
- `icon_intake_healpix`: ICON through Intake-ESM on HEALPix grid.
- `jra55`: JRA-55.
- `merra2`: Modern-Era Retrospective Analysis for Research and Applications, Version 2 (MERRA-2).
- `nzcsm`: New Zealand Convection Scale Model (NZCSM).
- `nzesm`: New Zealand Earth System Model (NZESM). [Experimental]
- `um`: UK Met Office Unified Model (UM).

Time format
-----------

`YYYY-MM-DD[THH:MM[:SS]]`, where `YYYY` is year, `MM` is month, `DD` is day, `HH` is hour, `MM` is minute, `SS` is second. Example: `2000-01-01T00:00:00`.

Files
-----

The track file is a NetCDF file containing 1D variables `lon`, `lat`, `time`, and optionally `time_bnds`. `time` and `time_bnds` are time in format conforming with the CF Conventions (has a valid `units` attribute and optional `calendar` attribute), `lon` is longitude between 0 and 360 degrees and `lat` is latitude between -90 and 90 degrees. If `time_bnds` is provided, discontinous track segments can be specified if adjacent time bounds are not coincident. The variables `lon`, `lat` and `time` have a single dimension `time`. The variable `time_bnds` has dimensions (`time`, `bnds`).

Examples
--------

Extract MERRA-2 model data in `M2I3NVASM.5.12.4` at 45 S, 170 E between 1 and 2 January 2020 and store the output in the directory `alcf_merra2_model`.

    alcf model merra2 point: { -45.0 170.0 } time: { 2020-01-01 2020-01-02 } M2I3NVASM.5.12.4 alcf_merra2_model
	'''
	d_track, time_lim = misc.cmd_point_or_track(point, time, track,
		track_gap=track_gap,
	)
	time_start = max(d_track['time_bnds'][0,0], time_lim[0])
	time_end = min(d_track['time_bnds'][-1,1], time_lim[1])

	model = MODELS.get(type_)
	if model is None:
		raise ValueError('Invalid type: %s' % type_)

	if njobs is None: njobs = os.cpu_count()

	warnings = []
	index = None
	if hasattr(model, 'index'):
		index = model.index(input_, warnings=warnings, recursive=r, njobs=njobs)

	with ProcessPoolExecutor(max_workers=njobs) as ex:
		fs = []
		tt = np.arange(np.floor(time_start - 0.5), np.ceil(time_end - 0.5)) + 0.5
		for t in tt:
			if not misc.track_has_seg(d_track, t, t + 1):
				continue
			f = ex.submit(worker, type_, input_, index, output, d_track, t,
				debug, r, override_year)
			fs += [f]
	for w in warnings:
		if len(w) == 2:
			print('Warning: %s' % w[0], file=sys.stderr)
			if debug: print(w[1], file=sys.stderr)
			else: print('Use --debug to print debugging information.', file=sys.stderr)
		else:
			print('Warning: %s' % w, file=sys.stderr)



#!/usr/bin/env python3

import signal
signal.signal(signal.SIGINT, signal.SIG_DFL)

import os
import sys
import pst
from alcf.cmds import main

def main_wrapper():
	args, kwargs = pst.decode_argv(sys.argv, as_unicode=True)
	ret = main.run(*args[1:], **kwargs)
	sys.exit(ret)

if __name__ == '__main__':
	main_wrapper()



from . import convert
from . import model
from . import simulate
from . import lidar
from . import stats
from . import compare
from . import plot
from . import calibrate
from . import auto
from . import download

CMDS = {
	'convert': convert.run,
	'model': model.run,
	'simulate': simulate.run,
	'lidar': lidar.run,
	'stats': stats.run,
	'compare': compare.run,
	'plot': plot.run,
	'calibrate': calibrate.run,
	'auto': auto.run,
	'download': download.run,
}



import copy
import numpy as np
from alcf import misc
import ds_format as ds

def output_sample(d, tres, output_sampling, epsilon=1/86400, align=True):
	t = d['time_bnds'][0,0]
	if align:
		r = (t + 0.5 + epsilon) % output_sampling
		t1 = t - r + epsilon
	else:
		t1 = t
	t2 = t1 + output_sampling

	dims = ds.get_dims(d)
	n = dims['time']
	n2 = int(output_sampling/tres)
	time = d['time']
	time_bnds = d['time_bnds']

	time_half2 = np.linspace(t1, t2, n2 + 1)
	time2 = 0.5*(time_half2[1:] + time_half2[:-1])

	for var in ds.get_vars(d):
		if 'time' not in d['.'][var]['.dims']:
			continue
		i = d['.'][var]['.dims'].index('time')
		x = d[var]
		size = x.shape
		size2 = list(size)
		size2[i] = n2
		x2 = np.full(size2, np.nan, dtype=x.dtype)
		for j in range(n2):
			mask = np.maximum(0,
				np.minimum(time_bnds[:,1], time_half2[j + 1]) -
				np.maximum(time_bnds[:,0], time_half2[j])
			)
			s = np.sum(mask)
			if s > 0.:
				mask /= np.sum(mask)
			sel2 = tuple([
				slice(None) if l != i else j
				for l in range(len(size))
			])
			for k in np.argwhere(mask > 0):
				sel = tuple([
					slice(None) if l != i else k
					for l in range(len(size))
				])
				x2[sel2] = np.where(np.isnan(x2[sel2])*mask[k], 0, x2[sel2]) + \
					x[sel]*mask[k]
		d[var] = x2
	d['time'] = time2
	d['time_bnds'] = np.full((n2, 2), np.nan, np.float64)
	d['time_bnds'][:,0] = time_half2[:-1]
	d['time_bnds'][:,1] = time_half2[1:]

def stream(dd, state,
	tres=None,
	tlim=None,
	output_sampling=None,
	align=True,
	**options,
):
	if tres is not None:
		state['aggregate_state'] = state.get('aggregate_state', {})
		dd = misc.aggregate(dd, state['aggregate_state'], output_sampling,
			align=align)
		return misc.stream(dd, state, output_sample,
			tres=tres,
			output_sampling=output_sampling,
			align=align,
		)
	return dd



import ds_format as ds
import os

def read(dirname, track, warnings=[]):
	pass
# 	time = track['time']
# 	t1, t2 = time[0], time[-1]
# 	d_cli = ds.read(dirname, ['cli'], {'time': [t1, t2]})
# 	d_clw = ds.read(dirname, ['clw'], {'time': [t1, t2]})
# 	ds.read(dirname)

# indexing
# - mask
# - list of indexes
# - list of values
# - index range
# - value range



import os
import sys
import numpy as np
import ds_format as ds
from alcf.algorithms import interp
from alcf.algorithms import stats
from alcf.misc import parse_time

VARIABLES = [
	'cloud_mask',
	'cbh',
	'zfull',
	'time',
	'time_bnds',
	'backscatter',
	'backscatter_sd',
	'backscatter_mol',
	'lon',
	'lat',
]

def read_filters(filters):
	if type(filters) in [list, tuple]:
		return [ds.read(filename) for filename in filters]
	else:
		return [ds.read(filters)]

def run(input_, output,
	tlim=None,
	blim=[5., 200.],
	bres=5.,
	bsd_lim=[0.001, 10.],
	bsd_log=True,
	bsd_res=0.001,
	bsd_z=8000.,
	filter=None,
	filter_exclude=None,
	filter_include=None,
	zlim=[0., 15000.],
	zres=100.,
	**kwargs
):
	'''
alcf-stats -- Calculate cloud occurrence statistics.
==========

Synopsis
--------

    alcf stats [<options>] [--] <input> <output>

Description
-----------

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `input`: Input filename or directory.
- `output`: Output filename or directory.

Options
-------

- `blim: <value>`: Backscatter histogram limits (10^-6 m-1.sr-1). Default: `{ 5 200 }`.
- `bres: <value>`: Backscatter histogram resolution (10^-6 m-1.sr-1). Default: `10`.
- `bsd_lim: { <low> <high> }`: Backscatter standard deviation histogram limits (10^-6 m-1.sr-1). Default: `{ 0.001 10 }`.
- `bsd_log: <value>`: Enable/disable logarithmic scale of the backscatter standard deviation histogram (`true` or `false`). Default: `true`.
- `bsd_res: <value>`: Backscatter standard deviation histogram resolution (10^-6 m-1.sr-1). Default: `0.001`.
- `bsd_z: <value>`: Backscatter standard deviation histogram height (m). Default: `8000`.
- `filter: <value> | { <value> ... }`: Filter profiles by condition: `cloudy` for cloudy profiles only, `clear` for clear sky profiles only, `night` for nighttime profiles, `day` for daytime profiles, `none` for all profiles. If an array of values is supplied, all conditions must be true. For `night` and `day`, lidar profiles must contain valid longitude and latitude fields set via the `lon` and `lat` arguments of `alcf lidar` or read implicitly from raw lidar data files if available (mpl, mpl2nc). Default: `none`.
- `filter_exclude: <value> | { <value>... }`: Filter by a mask defined in a NetCDF file, described below under Filter file. If multiple files are supplied, they must all apply for a profile to be excluded.
- `filter_include: <value> | { <value>... }`: The same as `filter_exclude`, but with time intervals to be included in the result. If both are defined, `filter_include` takes precedence. If multiple files are supplied, they must all apply for a profile to be included.
- `tlim: { <start> <end> }`: Time limits (see Time format below). Default: `none`.
- `zlim: { <low> <high> }`: Height limits (m). Default: `{ 0 15000 }`.
- `zres: <value>`: Height resolution (m). Default: `50`.

Time format
-----------

`YYYY-MM-DD[THH:MM[:SS]]`, where `YYYY` is year, `MM` is month, `DD` is day, `HH` is hour, `MM` is minute, `SS` is second. Example: `2000-01-01T00:00:00`.

Filter file
----------

The NetCDF file must define a variable `time_bnds` (float64), which are time intervals to be excluded from or included in the result. `time_bnds` must have two dimensions `time` of an arbitrary size and `bnds` of size 2. The first and second column of the variable should contain the start and end of the interval, respectively. `time_bnds` must be valid time in accordance with the CF Conventions.

Examples
--------

Calculate statistics from processed lidar data in `alcf_cl51_lidar` and store the output in `alcf_cl51_stats.nc`.

    alcf stats alcf_cl51_lidar alcf_cl51_stats.nc
	'''
	tlim_jd = parse_time(tlim) if tlim is not None else None
	state = {}
	options = {
		'tlim': tlim_jd,
		'blim': np.array(blim, dtype=np.float64)*1e-6,
		'bres': bres*1e-6,
		'bsd_lim': np.array(bsd_lim, dtype=np.float64)*1e-6,
		'bsd_log': bsd_log,
		'bsd_res': bsd_res*1e-6,
		'bsd_z': bsd_z,
		'filter': filter if type(filter) is list else [filter],
		'zlim': zlim,
		'zres': zres,
	}

	if filter_exclude is not None:
		dd = read_filters(filter_exclude)
		options['filters_exclude'] = [d['time_bnds'] for d in dd]

	if filter_include is not None:
		dd = read_filters(filter_include)
		options['filters_include'] = [d['time_bnds'] for d in dd]

	if os.path.isdir(input_):
		files = sorted(os.listdir(input_))
		for file_ in files:
			filename = os.path.join(input_, file_)
			if not os.path.isfile(filename):
				continue
			d = ds.read(filename, VARIABLES)
			print('<- %s' % filename)
			dd = stats.stream([d], state, **options)
	else:
		d = ds.read(input_, VARIABLES)
		print('<- %s' % input_)
		dd = stats.stream([d], state, **options)
	dd = stats.stream([None], state, **options)
	print('-> %s' % output)
	ds.write(output, dd[0])



import sys
import os
from alcf.cmds.auto_cmds import CMDS

def run(cmd, *args, **kwargs):
	'''
alcf-auto -- Peform automatic processing of model or lidar data.
=========

Synopsis
--------

    alcf auto model <model_type> <lidar_type> point: { <lon> <lat> } time: { <start> <end> } [<options>] [<model_options>] [<lidar_options>] [--] <input> <output>

    alcf auto model <model_type> <lidar_type> track: <track> [<options>] [<model_options>] [<lidar_options>] [--] <input> <output>

    alcf auto lidar <lidar_type> [<options>] [<lidar_options>] [--] <input> <output>

Description
-----------

`alcf auto model` is equivalent to the sequence of commands:

    alcf model
    alcf simulate
    alcf lidar
    alcf stats
    alcf stats (fine-scale)
    alcf stats (clear-sky fine-scale)
    alcf plot backscatter
    alcf plot backscatter_hist
    alcf plot backscatter_hist (fine-scale)
    alcf plot backscatter_hist (clear-sky fine-scale)
    alcf plot cloud_occurrence
    alcf plot cbh

`alcf auto lidar` is equivalent to the sequence of commands:

    alcf lidar
    alcf stats
    alcf stats (fine-scale)
    alcf stats (clear-sky fine-scale)
    alcf plot backscatter
    alcf plot backscatter_hist
    alcf plot backscatter_hist (fine-scale)
    alcf plot backscatter_hist (clear-sky fine-scale)
    alcf plot cloud_occurrence
    alcf plot cbh

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `end`: End time (see Time format below).
- `input`: Input directory containing model or lidar data, or, in case of `alcf auto compare`, the output of `alcf auto model` or `alcf auto lidar`.
- `lat`: Point latitutde.
- `lidar_options`: See `alcf lidar` options.
- `lidar_type`: Lidar type (see Lidar types below).
- `lon`: Point longitude.
- `model_options`: See `alcf model` options.
- `model_type`: Model type (see Model types below).
- `options`: See Options below.
- `plot_options`: See `alcf plot` options.
- `start`: Start time (see Time format below).
- `track`: Track NetCDF file (see Track below).

Options
-------

- `skip: <step>`: Skip all processing steps before `<step>`. `<step>` is one of: `model`, `simulate`, `lidar`, `stats`, `plot`. Default: `none`.

Model types
-----------

- `amps`: Antarctic Mesoscale Prediction System (AMPS).
- `era5`: ERA5.
- `icon`: ICON.
- `icon_intake_healpix`: ICON through Intake-ESM on HEALPix grid.
- `jra55`: JRA-55.
- `merra2`: Modern-Era Retrospective Analysis for Research and Applications, Version 2 (MERRA-2).
- `nzcsm`: New Zealand Convection Scale Model (NZCSM).
- `nzesm`: New Zealand Earth System Model (NZESM) (experimental).
- `um`: UK Met Office Unified Model (UM).

Lidar types
-----------

- `blview`: Vaisala BL-VIEW L2 product.
- `caliop`: CALIPSO/CALIOP (`alcf auto model` only).
- `chm15k`: Lufft CHM 15k.
- `cl31`: Vaisala CL31.
- `cl51`: Vaisala CL51.
- `cl61`: Vaisala CL61.
- `cosp`: COSP simulated lidar.
- `minimpl`: Sigma Space MiniMPL (converted via SigmaMPL).
- `mpl`: Sigma Space MPL (converted via SigmaMPL).
- `mpl2nc`: Sigma Space MPL and MiniMPL (converted via mpl2nc).

Time format
-----------

`YYYY-MM-DD[THH:MM[:SS]]`, where `YYYY` is year, `MM` is month, `DD` is day,
`HH` is hour, `MM` is minute, `SS` is second. Example: 2000-01-01T00:00:00.

Examples
--------

Simulate a Vaisala CL51 instrument from MERRA-2 data in `M2I3NVASM.5.12.4`
at 45 S, 170 E between 1 and 2 January 2020 and store the output in
`alcf_merra2`.

    alcf auto model merra2 cl51 point: { -45.0 170.0 } time: { 2020-01-01 2020-01-02 } M2I3NVASM.5.12.4 alcf_merra2

Process Lufft CHM 15k data in `chm15k` and store the output in `alcf_chm15k`.

    alcf auto lidar chm15k chm15k_data alcf_chm15k
	'''
	if cmd is None:
		sys.stderr.write(main.__doc__.strip() + '\n')
		return 1

	func = CMDS.get(cmd)
	if func is None:
		raise ValueError('Invalid command: %s' % cmd)
	func(*args, **kwargs)



META = {
	'time': {
		'.dims': ['time'],
		'long_name': 'time',
		'standard_name': 'time',
		'units': 'days since -4713-11-24 12:00 UTC',
		'calendar': 'proleptic_gregorian',
	},
	'time_bnds': {
		'.dims': ['time', 'bnds'],
		'long_name': 'time bounds',
		'standard_name': 'time',
		'units': 'days since -4713-11-24 12:00 UTC',
		'calendar': 'proleptic_gregorian',
	},
	'zfull': {
		'.dims': ['time', 'level'],
		'long_name': 'altitude of full-levels',
		'standard_name': 'altitude',
		'units': 'm',
	},
	'backscatter': {
		'.dims': ['time', 'level'],
		'long_name': 'total attenuated volume backscattering coefficient',
		'units': 'm-1 sr-1',
	},
	'altitude': {
		'.dims': ['time'],
		'long_name': 'instrument altitude',
		'standard_name': 'altitude',
		'units': 'm',
	},
	'lon': {
		'.dims': ['time'],
		'long_name': 'instrument longitude',
		'standard_name': 'longitude',
		'units': 'degrees_east',
	},
	'lat': {
		'.dims': ['time'],
		'long_name': 'instrument latitude',
		'standard_name': 'latitude',
		'units': 'degrees_north',
	},
}

from . import chm15k
from . import cl61
from . import blview
from . import cl51
from . import cl31
from . import mpl
from . import mpl2nc
from . import default
from . import caliop

LIDARS = {
	'default': default,
	'chm15k': chm15k,
	'cl61': cl61,
	'blview': blview,
	'cl51': cl51,
	'cl31': cl31,
	'mpl': mpl,
	'mpl2nc': mpl2nc,
	'minimpl': mpl,
	'cosp': default,
	'caliop': caliop,
}



import numpy as np
import ds_format as ds
from alcf import misc
from alcf.lidars import META

WAVELENGTH = 910 # nm
CALIBRATION_COEFF = 1e-9
SURFACE_LIDAR = True
SC_LR = 18.8 # sr. Stratocumulus lidar ratio (O'Connor et al., 2004).
MAX_RANGE = 15400 # m

VARS = {
	'backscatter': ['profile_data'],
	'zfull': ['range', 'altitude'],
	'altitude': ['altitude'],
}

DEFAULT_VARS = [
	'time',
]

def read(
	filename,
	vars,
	altitude=None,
	lon=None,
	lat=None,
	tlim=None,
	keep_vars=[],
	**kwargs
):
	sel = None
	tres = None
	if tlim is not None:
		d = ds.read(filename, 'time')
		d['time'] = d['time']/86400. + 2440587.5
		tres = d['time'][1] - d['time'][0]
		d['time_bnds'] = misc.time_bnds(d['time'], tres)
		mask = misc.time_mask(d['time_bnds'], tlim[0], tlim[1])
		if np.sum(mask) == 0: return None
		sel = { 'timeDim': mask }

	dep_vars = misc.dep_vars(VARS, vars)
	req_vars = dep_vars + DEFAULT_VARS + keep_vars
	d = ds.read(filename, req_vars, sel=sel, full=True)
	dx = {}
	misc.populate_meta(dx, META, set(vars) & set(VARS))
	n = ds.dim(d, 'timeDim')
	m = ds.dim(d, 'range')
	d['time'] = d['time']/86400. + 2440587.5
	if 'time' in vars:
		dx['time'] = d['time']
	if 'time_bnds' in vars:
		if tres is None: tres = d['time'][1] - d['time'][0]
		args = [] if tlim is None else [tlim[0], tlim[1]]
		dx['time_bnds'] = misc.time_bnds(d['time'], tres, *args)
	if 'altitude' in vars:
		dx['altitude'] = d['altitude'][:,0].astype(np.float64).filled(np.nan)
	if 'zfull' in vars:
		range_ = d['range'].astype(np.float64).filled(np.nan)
		dx['zfull'] = np.tile(range_, (n, 1))
		dx['zfull'] = (dx['zfull'].T + dx['altitude']).T
	if 'lon' in vars:
		dx['lon'] = np.full(n, lon, np.float64)
	if 'lat' in vars:
		dx['lat'] = np.full(n, lat, np.float64)
	if 'backscatter' in vars:
		profile_data = d['profile_data'].astype(np.float64).filled(np.nan)
		dx['backscatter'] = profile_data*CALIBRATION_COEFF
		j = np.max(np.nonzero(np.any(np.isfinite(dx['backscatter']), axis=0)))
		dx['backscatter'] = dx['backscatter'][:,:(j + 1)]
		if 'zfull' in vars:
			dx['zfull'] = dx['zfull'][:,:(j + 1)]
	for var in keep_vars:
		misc.keep_var(var, d, dx, {'time': 'timeDim', 'level': 'range'})
	return dx



import numpy as np
from alcf import misc

def lidar_ratio(d):
	dz = np.diff([0] + list(d['zfull']))
	if len(d['backscatter'].shape) == 3:
		n, m, l = d['backscatter'].shape
		bint = np.zeros((len(d['time']), l), dtype=np.float64)
		dims = ['time', 'column']
		for j in range(l):
			for i in range(len(d['time'])):
		 		bint[i,j] = np.sum(d['backscatter'][i,:,j]*dz)
	else:
		n, m = d['backscatter'].shape
		bint = np.zeros(n, dtype=np.float64)
		dims = ['time']
		for i in range(n):
		 	bint[i] = np.sum(d['backscatter'][i,:]*dz)
	with np.errstate(divide='ignore'):
		d['lr'] = 1./(2.*bint) # See O'Connor et al. (2004), Equation 6.
	d['.']['lr'] = {
		'.dims': dims,
		'long_name': 'effective lidar ratio',
		'units': 'sr',
	}

def stream(dd, state, **options):
	return misc.stream(dd, state, lidar_ratio, **options)



import numpy as np
import ds_format as ds
import datetime as dt
from alcf.lidars import META
from alcf.lidars import cl51

WAVELENGTH = 910 # nm
CALIBRATION_COEFF = 1.45e-3
SURFACE_LIDAR = True
SC_LR = 18.8 # sr. Stratocumulus lidar ratio (O'Connor et al., 2004).
MAX_RANGE = 7700 # m

def read(filename, vars, **kwargs):
	return cl51.read(filename, vars,
		calibration_coeff=CALIBRATION_COEFF,
		**kwargs
	)




META = {
	'clw': {
		'.dims': ['time', 'level'],
		'long_name': 'mass fraction of cloud liquid water',
		'standard_name': 'mass_fraction_of_cloud_liquid_water_in_air',
		'units': '1',
	},
	'cli': {
		'.dims': ['time', 'level'],
		'long_name': 'mass fraction of cloud ice',
		'standard_name': 'mass_fraction_of_cloud_ice_in_air',
		'units': '1',
	},
	'ps': {
		'.dims': ['time'],
		'long_name': 'surface air pressure',
		'standard_name': 'surface_air_pressure',
		'units': 'Pa',
	},
	'pfull': {
		'.dims': ['time', 'level'],
		'long_name': 'pressure at model full-levels',
		'standard_name': 'air_pressure',
		'units': 'Pa',
	},
	'zfull': {
		'.dims': ['time', 'level'],
		'long_name': 'altitude of model full-levels',
		'standard_name': 'height_above_reference_ellipsoid',
		'units': 'm',
	},
	'time': {
		'.dims': ['time'],
		'long_name': 'time',
		'standard_name': 'time',
		'units': 'days since -4713-11-24 12:00 UTC',
		'calendar': 'proleptic_gregorian',
	},
	'time_bnds': {
		'.dims': ['time', 'bnds'],
		'long_name': 'time bounds',
		'standard_name': 'time_bounds',
		'units': 'days since -4713-11-24 12:00 UTC',
		'calendar': 'proleptic_gregorian',
	},
	'lon': {
		'.dims': ['time'],
		'long_name': 'longitude',
		'standard_name': 'longitude',
		'units': 'degrees_east',
	},
	'lat': {
		'.dims': ['time'],
		'long_name': 'latitude',
		'standard_name': 'latitude',
		'units': 'degrees_north',
	},
	'ta': {
		'.dims': ['time', 'level'],
		'long_name': 'air temperature',
		'standard_name': 'air_temperature',
		'units': 'K',
	},
	'cl': {
		'.dims': ['time', 'level'],
		'long_name': 'cloud area fraction',
		'standard_name': 'cloud_area_fraction_in_atmosphere_layer',
		'units': '%',
	},
	'orog': {
		'.dims': ['time'],
		'long_name': 'surface altitude',
		'standard_name': 'surface_altitude',
		'units': 'm',
	},
}

from . import amps
#from . import cmip5
from . import era5
from . import jra55
from . import merra2
from . import nzcsm
from . import nzesm
from . import um
from . import icon
from . import icon_intake_healpix

MODELS = {
	'amps': amps,
	#'cmip5': cmip5,
	'era5': era5,
	'jra55': jra55,
	'merra2': merra2,
	'nzcsm': nzcsm,
	'nzesm': nzesm,
	'um': um,
	'icon': icon,
	'icon_intake_healpix': icon_intake_healpix,
}



cimport cython
cimport numpy as np
import numpy as np

def interp(
	np.ndarray[double, ndim=1] xhalf not None,
	np.ndarray[double, ndim=1] y not None,
	np.ndarray[double, ndim=1] xhalf2 not None
):
	'''Interpolate y(x) on x2'''
	cdef long n = len(xhalf)
	cdef long n2 = len(xhalf2)
	cdef np.ndarray[double, ndim=1] y2 = np.zeros(n2-1, dtype=np.float64)
	cdef long i = 0
	cdef long i2 = 0
	cdef double dx = 0
	cdef double dx2 = 0
	while i < n-1 and xhalf[i] < xhalf2[0]:
		i += 1
	if xhalf[i] < xhalf2[0]:
		return y2
	if i > 0:
		i -= 1
	while i2 < n2-1 and xhalf2[i2] < xhalf[-1]:
		dx2 = 0
		while i < n-1 and xhalf[i+1] < xhalf2[i2+1]:
			dx = xhalf[i+1] - max(xhalf2[i2], xhalf[i])
			if dx < 0:
				raise ValueError(xhalf[i+1], xhalf2[i2], xhalf[i])
			y2[i2] += y[i]*dx
			dx2 += dx
			i += 1
		if i < n-1 and xhalf[i] < xhalf2[i2+1]:
			dx = xhalf2[i2+1] - max(xhalf2[i2], xhalf[i])
			if dx < 0:
				raise ValueError(i, i2)
			y2[i2] += y[i]*dx
			dx2 += dx
		if dx2 > 0:
			y2[i2] /= dx2
		i2 += 1
	return y2



import os
import tempfile
import shutil
import numpy as np
import aquarius_time as aq
from alcf import misc
from alcf.download import MODELS

def run(type_,
	*args,
	point=None,
	time=None,
	track=None,
	login=False,
	overwrite=False,
	track_gap=21600,
	track_lon_180=False,
	nocache=False,
	**kwargs,
):
	'''
alcf-download -- Download model data.
=============

Synopsis
--------

    alcf download <type> [login_options] --login
    alcf download <type> point: { <lon> <lat> } time: { <start> <end> } [<options>] [--] <output>
    alcf download <type> track: <track> [<options>] [--] <output>

Description
-----------

This command downloads model data required by the lidar simulator for a given geographical point or ship track. Not all models supported by the other ALCF commands are supported by `alcf download` - these have to be acquired manually. Before downloading data, you have register an account on Copernicus CDS (ERA5; https://cds.climate.copernicus.eu) or NASA Earthdata (MERRA-2; https://earthdata.nasa.gov) and run `alcf download <type> --login`. This prompts for credentials interactively. For non-interactive use, use the `--overwrite` option and credential options (see the login options below). Only geographical subsets necessary to cover the point or track are downloaded.

Arguments following `--` are treated as literal strings. Use this delimiter if the output file name might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `type`: Model type (see Model types below).
- `lon`: Point longitude (degrees East).
- `lat`: Point latitutde (degrees North).
- `start`: Start time (see Time format below).
- `end`: End time (see Time format below).
- `track: <file>`, `track: { <file>... }`: One or more track NetCDF files (see Files below). If multiple files are supplied and `time_bnds` is not present in the files, they are assumed to be multiple segments of a discontinous track unless the last and first time of adjacent tracks are the same.
- `output`: Output directory.
- `login_options`: See login options below.
- `options`: See Options below.

Options
-------

- `--nocache`: Disable server-side caching of requests, when applicable.
- `--overwrite`: Overwrite existing files.
- `track_gap: <interval>`: If the interval is not 0, a track file is supplied, the `time_bnds` variable is not defined in the file and any two adjacent points are separated by more than the specified time interval (seconds), then a gap is assumed to be present between the two data points, instead of interpolating location between the two points. Default: `21600` (6 hours).

Login options
-------------

- `--overwrite`: Overwrite existing files.

MERRA-2 login options
---------------------

- `user: <value>`: Supply non-interactive user instead of prompting.
- `password: <value>`: Supply non-interactive password instead of prompting.

ERA5 login options
------------------

- `uid: <value>`: Supply non-interactive UID instead of prompting.
- `key: <value>`: Supply non-interactive API key instead of prompting.

Model types
-----------

- `era5`: ERA5.
- `merra2`: MERRA-2.

Files
-----

The track file is a NetCDF file containing 1D variables `lon`, `lat`, `time`, and optionally `time_bnds`. `time` and `time_bnds` are time in format conforming with the CF Conventions (has a valid `units` attribute and optional `calendar` attribute), `lon` is longitude between 0 and 360 degrees and `lat` is latitude between -90 and 90 degrees. If `time_bnds` is provided, discontinous track segments can be specified if adjacent time bounds are not coincident. The variables `lon`, `lat` and `time` have a single dimension `time`. The variable `time_bnds` has dimensions (`time`, `bnds`).

Time format
-----------

`YYYY-MM-DD[THH:MM[:SS]]`, where `YYYY` is year, `MM` is month, `DD` is day, `HH` is hour, `MM` is minute, `SS` is second. Example: `2000-01-01T00:00:00`.

Examples
--------

Log in to the Copernicus CDS (ERA5) service:

    alcf download era5 --login

Download ERA5 data at 45 S, 170 E between 1 and 2 January 2020 and store the output in the directory `era5`.

    alcf download era5 point: { -45.0 170.0 } time: { 2020-01-01 2020-01-02 } era5

Log in to the NASA Earthdata (MERRA-2) service:

    alcf download merra2 --login

Download MERRA-2 data for a ship track `track.nc` and store the output in the directory `merra2`.

    alcf download merra2 track: track.nc merra2
	'''
	model = MODELS.get(type_)
	if model is None:
		raise ValueError('Invalid type: %s' % type_)

	if login:
		model.login(overwrite=overwrite, **kwargs)
		return

	if len(args) != 1:
		raise TypeError('run() ')
	output = args[0]

	d, time_lim = misc.cmd_point_or_track(point, time, track,
		track_gap=track_gap,
	)
	time_start = max(d['time_bnds'][0,0], time_lim[0])
	time_end = min(d['time_bnds'][-1,1], time_lim[1])

	try: os.mkdir(output)
	except FileExistsError: pass

	with tempfile.TemporaryDirectory() as tmpdir:
		tt = np.arange(np.floor(time_start - 0.5), np.ceil(time_end - 0.5)) + 0.5
		for t in tt:
			t1 = t
			t2 = t + 1
			if not misc.track_has_seg(d, t1, t2):
				continue
			mask = ((d['time_bnds'][:,0] <= t2) & (d['time_bnds'][:,1] >= t1))
			lon = d['lon'][mask]
			lat = d['lat'][mask]
			lon1 = np.floor(np.min(lon))
			lon2 = np.ceil(np.max(lon))
			lat1 = np.floor(np.min(lat))
			lat2 = np.ceil(np.max(lat))
			if lon1 == lon2:
				lon1 -= 1
				lon2 += 1
			if lat1 == lat2:
				lat1 -= 1
				lat2 += 1
			for lon12 in [lon1, lon2]:
				if lon12 < 0: lon12 += 360
				if lon12 > 360: lon12 -= 360
			for lat12 in [lat1, lat2]:
				if lat12 < -90: lat12 = -90
				if lat12 > 90: lat12 = 90
			date = aq.to_date(t)
			year, month, day = date[1:4]
			name = '%04d-%02d-%02d.nc' % (year, month, day)
			for product in model.PRODUCTS:
				try: os.mkdir(os.path.join(tmpdir, product))
				except FileExistsError: pass
				if len(model.PRODUCTS) == 1:
					file_ = name
				else:
					file_ = os.path.join(product, name)
					try: os.mkdir(os.path.join(output, product))
					except FileExistsError: pass
				outfile = os.path.join(output, file_)
				if os.path.exists(outfile) and not overwrite:
					continue
				filename = os.path.join(tmpdir, file_)
				model.download(filename, product,
					year, month, day,
					lon1, lon2, lat1, lat2,
					nocache=nocache)
				shutil.move(filename, outfile)
				print('-> %s' % outfile)



import sys
import re
from alcf.cmds import CMDS

def md_to_text(md):
	text = md.strip()
	text += '\n\nBug reporting\n-------------\n\nReport bugs to Peter Kuma (peter@peterkuma.net).\n'
	return text

def run(cmd=None, *args, **kwargs):
	'''alcf -- Tool for processing of automatic lidar and ceilometer (ALC) data and intercomparison with atmospheric models.
====

Synopsis
--------

    alcf <cmd> [<options>]
    alcf <cmd> --help

Arguments
---------

- `cmd`: See Commands below.
- `options`: Command options.

Commands
--------

- `auto`: Peform automatic processing of model or lidar data.
- `calibrate`: Calibrate lidar backscatter.
- `convert`: Convert input instrument or model data to the ALCF standard NetCDF.
- `download`: Download model data.
- `lidar`: Process lidar data.
- `model`: Extract model data at a point or along a track.
- `plot`: Plot lidar data.
- `simulate`: Simulate lidar measurements from model data using COSP.
- `stats`: Calculate cloud occurrence statistics.

Options
-------

- `--help`: Print help for command.
- `--debug`: Enable debugging information.
'''

	if cmd is None:
		sys.stderr.write(md_to_text(run.__doc__))
		return 1

	func = CMDS.get(cmd)
	if func is None:
		raise ValueError('Invalid command: %s' % cmd)
	try:
		func(*args, **kwargs)
	except TypeError as e:
		if str(e).startswith('run() '):
			sys.stderr.write(md_to_text(func.__doc__))
			return 1
		else:
			raise e
	return 0
